"""For testing netneurotools.plotting.pysurfer_plotters functionality."""

import pytest
import numpy as np
from netneurotools import datasets, plotting


@pytest.mark.filterwarnings('ignore')
def test_plot_fsvertex():
    """Test plotting on a freesurfer vertex."""
    surfer = pytest.importorskip('surfer')

    data = np.random.rand(20484)
    brain = plotting.plot_fsvertex(data, subject_id='fsaverage5',
                                   offscreen=True)
    assert isinstance(brain, surfer.Brain)


@pytest.mark.filterwarnings('ignore')
def test_plot_fsaverage():
    """Test plotting on a freesurfer average brain."""
    surfer = pytest.importorskip('surfer')

    data = np.random.rand(68)
    lhannot, rhannot = datasets.fetch_cammoun2012('fsaverage5')['scale033']
    brain = plotting.plot_fsaverage(data, lhannot=lhannot, rhannot=rhannot,
                                    subject_id='fsaverage5', offscreen=True)
    assert isinstance(brain, surfer.Brain)



"""Functions for constucting networks."""


from .consensus import (
    func_consensus, struct_consensus
)


from .randomize import (
    randmio_und,
    match_length_degree_distribution,
    strength_preserving_rand_sa,
    strength_preserving_rand_sa_mse_opt,
    strength_preserving_rand_sa_dir
)


from .networks_utils import (
    binarize_network, threshold_network, get_triu
)


__all__ = [
    # consensus
    'func_consensus', 'struct_consensus',
    # generative
    # randomize
    'randmio_und', 'match_length_degree_distribution',
    'strength_preserving_rand_sa', 'strength_preserving_rand_sa_mse_opt',
    'strength_preserving_rand_sa_dir',
    # networks_utils
    'binarize_network', 'threshold_network', 'get_triu'
]



"""Functions for calculating correlation."""

import numpy as np
import scipy.stats as sstats
import scipy.special as sspecial
from sklearn.utils.validation import check_random_state

try:  # scipy >= 1.8.0
    from scipy.stats._stats_py import _chk2_asarray
except ImportError:  # scipy < 1.8.0
    from scipy.stats.stats import _chk2_asarray


from .. import has_numba

if has_numba:
    from numba import njit


def efficient_pearsonr(a, b, ddof=1, nan_policy="propagate"):
    """
    Compute correlation of matching columns in `a` and `b`.

    Parameters
    ----------
    a,b : array_like
        Sample observations. These arrays must have the same length and either
        an equivalent number of columns or be broadcastable
    ddof : int, optional
        Degrees of freedom correction in the calculation of the standard
        deviation. Default: 1
    nan_policy : bool, optional
        Defines how to handle when input contains nan. 'propagate' returns nan,
        'raise' throws an error, 'omit' performs the calculations ignoring nan
        values. Default: 'propagate'

    Returns
    -------
    corr : float or numpy.ndarray
        Pearson's correlation coefficient between matching columns of inputs
    pval : float or numpy.ndarray
        Two-tailed p-values

    Notes
    -----
    If either input contains nan and nan_policy is set to 'omit', both arrays
    will be masked to omit the nan entries.

    Examples
    --------
    >>> from netneurotools import stats

    Generate some not-very-correlated and some highly-correlated data:

    >>> np.random.seed(12345678)  # set random seed for reproducible results
    >>> x1, y1 = stats.make_correlated_xy(corr=0.1, size=100)
    >>> x2, y2 = stats.make_correlated_xy(corr=0.8, size=100)

    Calculate both correlations simultaneously:

    >>> stats.efficient_pearsonr(np.c_[x1, x2], np.c_[y1, y2])
    (array([0.10032565, 0.79961189]), array([3.20636135e-01, 1.97429944e-23]))
    """
    a, b, _ = _chk2_asarray(a, b, 0)
    if len(a) != len(b):
        raise ValueError("Provided arrays do not have same length")

    if a.size == 0 or b.size == 0:
        return np.nan, np.nan

    if nan_policy not in ("propagate", "raise", "omit"):
        raise ValueError(f'Value for nan_policy "{nan_policy}" not allowed')

    a, b = a.reshape(len(a), -1), b.reshape(len(b), -1)
    if a.shape[1] != b.shape[1]:
        a, b = np.broadcast_arrays(a, b)

    mask = np.logical_or(np.isnan(a), np.isnan(b))
    if nan_policy == "raise" and np.any(mask):
        raise ValueError('Input cannot contain NaN when nan_policy is "omit"')
    elif nan_policy == "omit":
        # avoid making copies of the data, if possible
        a = np.ma.masked_array(a, mask, copy=False, fill_value=np.nan)
        b = np.ma.masked_array(b, mask, copy=False, fill_value=np.nan)

    with np.errstate(invalid="ignore"):
        corr = sstats.zscore(a, ddof=ddof, nan_policy=nan_policy) * sstats.zscore(
            b, ddof=ddof, nan_policy=nan_policy
        )

    sumfunc, n_obs = np.sum, len(a)
    if nan_policy == "omit":
        corr = corr.filled(np.nan)
        sumfunc = np.nansum
        n_obs = np.squeeze(np.sum(np.logical_not(np.isnan(corr)), axis=0))

    corr = sumfunc(corr, axis=0) / (n_obs - 1)
    corr = np.squeeze(np.clip(corr, -1, 1)) / 1

    # taken from scipy.stats
    ab = (n_obs / 2) - 1
    prob = 2 * sspecial.betainc(ab, ab, 0.5 * (1 - np.abs(corr)))

    return corr, prob


def fast_pearsonr():
    """Calculate Pearson correlation coefficient."""
    pass


def _weighted_mean(x, w):
    return np.sum(x * w) / np.sum(w)


if has_numba:
    _weighted_mean = njit(_weighted_mean)


def _weighted_pearsonr_vectorized(x_vec, y_vec, weight_vec):
    x_bar_diff = x_vec - _weighted_mean(x_vec, weight_vec)
    y_bar_diff = y_vec - _weighted_mean(y_vec, weight_vec)
    weight_vec_sum = np.sum(weight_vec)
    cov_x_y = np.sum(weight_vec * x_bar_diff * y_bar_diff) / weight_vec_sum
    cov_x_x = np.sum(weight_vec * x_bar_diff * x_bar_diff) / weight_vec_sum
    cov_y_y = np.sum(weight_vec * y_bar_diff * y_bar_diff) / weight_vec_sum
    return cov_x_y / np.sqrt(cov_x_x * cov_y_y)


def _weighted_pearsonr_numba(x_vec, y_vec, weight_vec):
    n = len(x_vec)
    x_weighted_mean = _weighted_mean(x_vec, weight_vec)
    y_weighted_mean = _weighted_mean(y_vec, weight_vec)
    upper, lower1, lower2 = 0, 0, 0
    for i in range(n):
        upper += (
            weight_vec[i] * (x_vec[i] - x_weighted_mean) * (y_vec[i] - y_weighted_mean)
        )
        lower1 += weight_vec[i] * (x_vec[i] - x_weighted_mean) ** 2
        lower2 += weight_vec[i] * (y_vec[i] - y_weighted_mean) ** 2
    return upper / np.sqrt(lower1 * lower2)


if has_numba:
    _weighted_pearsonr_numba = njit(_weighted_pearsonr_numba)


def weighted_pearsonr(x_vec, y_vec, weight_vec, use_numba=has_numba):
    r"""
    Calculate weighted Pearson correlation coefficient.

    Parameters
    ----------
    x_vec : array_like
        First vector of data
    y_vec : array_like
        Second vector of data
    weight_vec : array_like
        Vector of weights
    use_numba : bool, optional
        Whether to use numba for calculation. Default: True
        (if numba is available).

    Returns
    -------
    corr : float
        Weighted Pearson correlation coefficient

    Notes
    -----
    This function calculates the weighted Pearson correlation coefficient between
    two vectors, defined as:

    .. math::
        r = \frac{\sum_i w_i (x_i - \bar{x})(y_i - \bar{y})}
                    {\sqrt{\sum_i w_i (x_i - \bar{x})^2 \sum_i w_i (y_i - \bar{y})^2}}

    where :math:`x_i` and :math:`y_i` are the data points, :math:`w_i` are the
    weights, and :math:`\bar{x}` and :math:`\bar{y}` are the weighted means of
    the data points.

    """
    if use_numba:
        if not has_numba:
            raise ValueError("Numba not installed; cannot use numba for calculation")
        return _weighted_pearsonr_numba(x_vec, y_vec, weight_vec)
    else:
        return _weighted_pearsonr_vectorized(x_vec, y_vec, weight_vec)


def make_correlated_xy(corr=0.85, size=10000, seed=None, tol=0.001):
    """
    Generate random vectors that are correlated to approximately `corr`.

    Parameters
    ----------
    corr : [-1, 1] float or (N, N) numpy.ndarray, optional
        The approximate correlation desired. If a float is provided, two
        vectors with the specified level of correlation will be generated. If
        an array is provided, it is assumed to be a symmetrical correlation
        matrix and ``len(corr)`` vectors with the specified levels of
        correlation will be generated. Default: 0.85
    size : int or tuple, optional
        Desired size of the generated vectors. Default: 1000
    seed : {int, np.random.RandomState instance, None}, optional
        Seed for random number generation. Default: None
    tol : [0, 1] float, optional
        Tolerance of correlation between generated `vectors` and specified
        `corr`. Default: 0.001

    Returns
    -------
    vectors : numpy.ndarray
        Random vectors of size `size` with correlation specified by `corr`

    Examples
    --------
    >>> from netneurotools import stats

    By default two vectors are generated with specified correlation

    >>> x, y = stats.make_correlated_xy()
    >>> np.corrcoef(x, y)  # doctest: +SKIP
    array([[1.        , 0.85083661],
           [0.85083661, 1.        ]])
    >>> x, y = stats.make_correlated_xy(corr=0.2)
    >>> np.corrcoef(x, y)  # doctest: +SKIP
    array([[1.        , 0.20069953],
           [0.20069953, 1.        ]])

    You can also provide correlation matrices to generate more than two vectors
    if desired. Note that this makes it more difficult to ensure the actual
    correlations are close to the desired values:

    >>> corr = [[1, 0.5, 0.3], [0.5, 1, 0], [0.3, 0, 1]]
    >>> out = stats.make_correlated_xy(corr=corr)
    >>> out.shape
    (3, 10000)
    >>> np.corrcoef(out)  # doctest: +SKIP
    array([[1.        , 0.50965273, 0.30235686],
           [0.50965273, 1.        , 0.01089107],
           [0.30235686, 0.01089107, 1.        ]])
    """
    rs = check_random_state(seed)

    # no correlations outside [-1, 1] bounds
    if np.any(np.abs(corr) > 1):
        raise ValueError("Provided `corr` must (all) be in range [-1, 1].")

    # if we're given a single number, assume two vectors are desired
    if isinstance(corr, (int, float)):
        covs = np.ones((2, 2)) * 0.111
        covs[(0, 1), (1, 0)] *= corr
    # if we're given a correlation matrix, assume `N` vectors are desired
    elif isinstance(corr, (list, np.ndarray)):
        corr = np.asarray(corr)
        if corr.ndim != 2 or len(corr) != len(corr.T):
            raise ValueError(
                "If `corr` is a list or array, must be a 2D "
                "square array, not {}".format(corr.shape)
            )
        if np.any(np.diag(corr) != 1):
            raise ValueError("Diagonal of `corr` must be 1.")
        covs = corr * 0.111
    means = [0] * len(covs)

    # generate the variables
    count = 0
    while count < 500:
        vectors = rs.multivariate_normal(mean=means, cov=covs, size=size).T
        flat = vectors.reshape(len(vectors), -1)
        # if diff between actual and desired correlations less than tol, break
        if np.all(np.abs(np.corrcoef(flat) - (covs / 0.111)) < tol):
            break
        count += 1

    return vectors



"""For testing netneurotools.plotting.color_utils functionality."""


def test_register_cmaps():
    """Test registering colormaps."""
    import matplotlib
    if "justine" in matplotlib.colormaps:
        assert True
    else:
        assert False



"""For testing netneurotools.plotting.pyvista_plotters functionality."""

import pytest


@pytest.mark.pyvista
def test_pyvista_smoke():
    """Test that pyvista is importable."""
    import pyvista  # noqa: F401



"""Functions for performing statistical operations."""


from .correlation import (
    efficient_pearsonr,
    weighted_pearsonr,
    make_correlated_xy
)


from .permutation_test import (
    permtest_1samp,
    permtest_rel,
    permtest_pearsonr
)


from .regression import (
    _add_constant,
    residualize,
    get_dominance_stats
)


# from .stats_utils import ()


__all__ = [
    # correlation
    'efficient_pearsonr', 'weighted_pearsonr', 'make_correlated_xy',
    # permutation_test
    'permtest_1samp', 'permtest_rel', 'permtest_pearsonr',
    # regression
    '_add_constant', 'residualize', 'get_dominance_stats',
    # stats_utils
]



"""Functions for working with network modularity."""


from .modules import (
    match_cluster_labels,
    match_assignments,
    reorder_assignments,
    find_consensus,
    consensus_modularity,
    _dummyvar,
    zrand,
    _zrand_partitions,
    get_modularity,
    get_modularity_z,
    get_modularity_sig,
)


__all__ = [
    # modules
    'match_cluster_labels', 'match_assignments', 'reorder_assignments',
    'find_consensus', 'consensus_modularity', '_dummyvar', 'zrand',
    '_zrand_partitions', 'get_modularity', 'get_modularity_z',
    'get_modularity_sig',
]



"""Utilites for loading / creating datasets."""

import json
import os
import shutil
from pathlib import Path
from collections import namedtuple
import importlib.resources


try:
    # nilearn 0.10.3
    from nilearn.datasets._utils import fetch_single_file as _fetch_file
except ImportError:
    from nilearn.datasets.utils import _fetch_file


SURFACE = namedtuple("Surface", ("L", "R"))

FREESURFER_IGNORE = [
    "unknown",
    "corpuscallosum",
    "Background+FreeSurfer_Defined_Medial_Wall",
]


def _get_data_dir(data_dir=None):
    """
    Get path to netneurotools data directory.

    Parameters
    ----------
    data_dir : str, optional
        Path to use as data directory. If not specified, will check for
        environmental variable 'NNT_DATA'; if that is not set, will use
        `~/nnt-data` instead. Default: None

    Returns
    -------
    data_dir : str
        Path to use as data directory
    """
    if data_dir is None:
        data_dir = os.environ.get("NNT_DATA", str(Path.home() / "nnt-data"))
    data_dir = Path(data_dir).expanduser()
    data_dir.mkdir(parents=True, exist_ok=True)
    return data_dir


def _decode_url(url_type, url):
    """
    Format `data` object with OSF API URL.

    Parameters
    ----------
    data : object
        If dict with a `url` key, will format OSF_API with relevant values

    Returns
    -------
    data : object
        Input data with all `url` dict keys formatted
    """
    OSF_API = "https://files.osf.io/v1/resources/{}/providers/osfstorage/{}"
    GITHUB_RELEASE_API = "https://github.com/{}/{}/archive/refs/tags/{}.tar.gz"
    GITHUB_ARCHIVE_API = "https://github.com/{}/{}/archive/{}.tar.gz"
    ZENODO_FILE_API = "https://zenodo.org/record/{}/files/{}?download=1"

    if url_type == "osf":
        out_url = OSF_API.format(*url)
    elif url_type == "github-release":
        out_url = GITHUB_RELEASE_API.format(*url)
    elif url_type == "github-archive":
        out_url = GITHUB_ARCHIVE_API.format(*url)
    elif url_type == "zenodo-file":
        out_url = ZENODO_FILE_API.format(*url)
    else:
        raise ValueError("URL type {} not recognized".format(url_type))

    return out_url


def fetch_file(dataset_name, keys=None, force=False, data_dir=None, verbose=1):
    """
    Fetch file(s) for dataset `dataset_name`.

    Parameters
    ----------
    dataset_name : str
        Name of dataset to fetch
    keys : str or list, optional
        If provided, will only fetch the specified key(s) from the dataset.
        Default: None
    force : bool, optional
        If True, will overwrite existing dataset. Default: False
    data_dir : str, optional
        Path to use as data directory. If not specified, will check for
        environmental variable 'NNT_DATA'; if that is not set, will use
        `~/nnt-data` instead. Default: None
    verbose : int, optional
        Verbosity level. Default: 1
    """
    data_dir = _get_data_dir(data_dir=data_dir)
    info = _get_dataset_info(dataset_name)

    # deal with nested keys
    if keys is not None:
        if isinstance(keys, str):
            keys = [keys]
        for k in keys:
            info = info[k]

    if "uncompress" in info and info["uncompress"]:
        targ_folder = data_dir / info["rel-path"]

        # check if folder exists and remove if force=True
        if targ_folder.exists():
            if force:
                shutil.rmtree(targ_folder)
                if verbose:
                    print(f"Dataset {dataset_name} already exists. Overwriting.")
            else:
                if verbose:
                    print(f"Dataset {dataset_name} already exists. Skipping download.")
                return targ_folder

        # download compressed file
        dl_fname = _fetch_file(
            _decode_url(info["url-type"], info["url"]),
            data_dir,
            resume=True,
            md5sum=info["md5"],
            verbose=verbose,
        )

        # extract contents and remove compressed file
        if info["url-type"] == "zenodo-file":
            archive_format = "zip"
        else:
            archive_format = "gztar"

        shutil.unpack_archive(dl_fname, targ_folder.parent, format=archive_format)
        os.remove(dl_fname)

        # rename folder if necessary
        if "rename-folder" in info:
            shutil.move(targ_folder.parent / info["rename-folder"], targ_folder)
        if verbose:
            print(f"Downloaded {dataset_name} to {data_dir}")
        return targ_folder
    else:
        targ_file = data_dir / dataset_name / info["file-name"]

        # check if file exists and remove if force=True
        if targ_file.exists():
            if force:
                os.remove(targ_file)
                if verbose:
                    print(f"Dataset {dataset_name} already exists. Overwriting.")
            else:
                if verbose:
                    print(f"Dataset {dataset_name} already exists. Skipping download.")
                return targ_file
        # download file
        dl_fname = _fetch_file(
            _decode_url(info["url-type"], info["url"]),
            data_dir / dataset_name,
            resume=True,
            md5sum=info["md5"],
            verbose=verbose,
        )
        # move/rename file
        shutil.move(dl_fname, targ_file)
        if verbose:
            print(f"Downloaded {dataset_name} to {data_dir}")
        return targ_file


def _load_resource_json(relative_path):
    """
    Load JSON file from package resources.

    Parameters
    ----------
    relative_path : str
        Path to JSON file relative to package resources

    Returns
    -------
    resource_json : dict
        JSON file loaded as a dictionary
    """
    # handling pkg_resources.resource_filename deprecation
    if getattr(importlib.resources, "files", None) is not None:
        f_resource = importlib.resources.files("netneurotools") / relative_path
    else:
        from pkg_resources import resource_filename

        f_resource = resource_filename("netneurotools", relative_path)

    with open(f_resource) as src:
        resource_json = json.load(src)

    return resource_json


NNT_DATASETS = _load_resource_json("datasets/datasets.json")
# NNT_DATASETS = _decode_urls(NNT_DATASETS)


def _get_dataset_info(name):
    """
    Return url and MD5 checksum for dataset `name`.

    Parameters
    ----------
    name : str
        Name of dataset

    Returns
    -------
    url : str
        URL from which to download dataset
    md5 : str
        MD5 checksum for file downloade from `url`
    """
    try:
        return NNT_DATASETS[name]
    except KeyError:
        raise KeyError(
            f"Provided dataset {name} is not valid. "
            f"Must be one of: {sorted(NNT_DATASETS.keys())}"
        ) from None


NNT_REFERENCES = _load_resource_json("datasets/references.json")


def _get_reference_info(name, verbose=1, return_dict=False):
    """
    Return reference information for dataset `name`.

    Parameters
    ----------
    name : str
        Name of dataset

    Returns
    -------
    reference : str
        Reference information for dataset
    """
    try:
        curr_refs = NNT_REFERENCES[name]
        if verbose:
            print("Please cite the following papers if you are using this function:")
            for bib_category, bib_category_items in curr_refs.items():
                print(f"  [{bib_category}]:")
                for bib_item in bib_category_items:
                    print(f"    {bib_item['citation']}")

        if return_dict:
            return curr_refs
    except KeyError:
        raise KeyError(
            f"Provided dataset {name} is not valid. "
            f"Must be one of: {sorted(NNT_REFERENCES.keys())}"
        ) from None


def _fill_reference_json(bib_file, json_file, overwrite=False, use_defaults=False):
    """
    Fill in citation information for references in a JSON file.

    For internal use only.

    Parameters
    ----------
    bib_file : str
        Path to BibTeX file containing references
    json_file : str
        Path to JSON file containing references
    overwrite : bool, optional
        Whether to overwrite existing citation information. Default: False
    use_defaults : bool, optional
        Whether to use default paths for `bib_file` and `json_file`. Default: False

    Returns
    -------
    None
    """
    if use_defaults:
        bib_file = (
            importlib.resources.files("netneurotools") / "datasets/netneurotools.bib"
        )
        json_file = (
            importlib.resources.files("netneurotools") / "datasets/references.json"
        )

    from pybtex import PybtexEngine

    engine = PybtexEngine()

    def _get_citation(key):
        s = engine.format_from_file(
            filename=bib_file,
            style="unsrt",
            citations=[key],
            output_backend="plaintext",
        )
        return s.strip("\n").replace("[1] ", "")

    with open(json_file) as src:
        nnt_refs = json.load(src)

    for _, value in nnt_refs.items():
        for bib_category in value:
            for bib_item in value[bib_category]:
                if bib_item["bibkey"] not in ["", None]:
                    if bib_item["citation"] == "" or overwrite:
                        bib_item["citation"] = _get_citation(bib_item["bibkey"])

    with open(json_file, "w") as dst:
        json.dump(nnt_refs, dst, indent=4)


def _check_freesurfer_subjid(subject_id, subjects_dir=None):
    """
    Check that `subject_id` exists in provided FreeSurfer `subjects_dir`.

    Parameters
    ----------
    subject_id : str
        FreeSurfer subject ID
    subjects_dir : str, optional
        Path to FreeSurfer subject directory. If not set, will inherit from
        the environmental variable $SUBJECTS_DIR. Default: None

    Returns
    -------
    subject_id : str
        FreeSurfer subject ID, as provided
    subjects_dir : str
        Full filepath to `subjects_dir`

    Raises
    ------
    FileNotFoundError
    """
    # check inputs for subjects_dir and subject_id
    if subjects_dir is None or not os.path.isdir(subjects_dir):
        try:
            subjects_dir = Path(os.environ["SUBJECTS_DIR"])
        except KeyError:
            subjects_dir = Path.cwd()
    else:
        subjects_dir = Path(subjects_dir).resolve()

    subjdir = subjects_dir / subject_id
    if not subjdir.is_dir():
        raise FileNotFoundError(
            f"Cannot find specified subject id {subject_id} in "
            f"provided subject directory {subjects_dir}."
        )

    return subject_id, subjects_dir


def _get_freesurfer_subjid(subject_id, subjects_dir=None):
    """
    Get fsaverage version `subject_id`, fetching if required.

    Parameters
    ----------
    subject_id : str
        FreeSurfer subject ID
    subjects_dir : str, optional
        Path to FreeSurfer subject directory. If not set, will inherit from
        the environmental variable $SUBJECTS_DIR. Default: None

    Returns
    -------
    subject_id : str
        FreeSurfer subject ID
    subjects_dir : str
        Path to subject directory with `subject_id`
    """
    # check for FreeSurfer install w/fsaverage; otherwise, fetch required
    try:
        subject_id, subjects_dir = _check_freesurfer_subjid(subject_id, subjects_dir)
    except FileNotFoundError:
        if "fsaverage" not in subject_id:
            raise ValueError(
                f"Provided subject {subject_id} does not exist in provided "
                f"subjects_dir {subjects_dir}"
            ) from None
        from .fetch_template import fetch_fsaverage

        fetch_fsaverage(subject_id)
        subjects_dir = os.path.join(_get_data_dir(), "tpl-fsaverage")
        subject_id, subjects_dir = _check_freesurfer_subjid(subject_id, subjects_dir)

    return subject_id, subjects_dir



"""Functions for working with network modules."""

import bct
import numpy as np
from sklearn.utils.validation import check_random_state
from scipy import optimize
from scipy.cluster import hierarchy

from .. import has_numba
if has_numba:
    from numba import njit


def _get_relabels(c1, c2):
    """
    Find mapping of labels from `c1` to `c2`.

    Parameters
    ----------
    c1, c2 : (N,) array_like
        Cluster labels for `N` subjects

    Returns
    -------
    src, tar : numpy.ndarray
        Source-target mapping of labels in `c1`
    """

    def _match(c1s, c2s):
        intersect = len(np.intersect1d(c1s, c2s))
        return (len(c1s) - intersect) + (len(c2s) - intersect)

    # get unique IDs of clusters in both solutions
    ids1, ids2 = np.unique(c1), np.unique(c2)

    idxs = np.arange(len(c1))
    assignments = np.ones((len(ids1), len(ids2)), dtype=int) * -1

    for n, i in enumerate(ids1):
        c1s = idxs[c1 == i]
        assignments[n] = [_match(c1s, idxs[c2 == f]) for f in ids2]

    idx1, idx2 = optimize.linear_sum_assignment(assignments)

    return ids1[idx1], ids2[idx2]


def match_cluster_labels(source, target):
    """
    Align cluster labels in `source` to those in `target`.

    Uses :func:`scipy.optimize.linear_sum_assignment` to match solutions. If
    `source` has fewer clusters than `target` the returned assignments may be
    discontinuous (see Examples for more information).

    Parameters
    ----------
    source : (N,) array_like
        Cluster labels for `N` subjects, to be re-labelled
    target : (N,) array_like
        Cluster labels for `N` subjects, to which `source` is mapped

    Returns
    -------
    matched : (N,) array_like
        Re-labelled `source` with cluster assignments "matched" to `target`

    Examples
    --------
    >>> from netneurotools import modularity

    When cluster labels are perfectly matched but e.g., inverted the function
    will find a perfect mapping:

    >>> a = np.array([1, 1, 1, 0, 0, 0, 0, 0, 0, 0])
    >>> b = np.array([0, 0, 0, 1, 1, 1, 1, 1, 1, 1])
    >>> modularity.match_cluster_labels(a, b)
    array([0, 0, 0, 1, 1, 1, 1, 1, 1, 1])

    However, the mapping will work even when cluster assignments between the
    two solutions aren't perfectly matched. The function will simply choose a
    re-labelling that generates the "best" alignment between labels:

    >>> a = np.array([0, 0, 0, 2, 2, 2, 2, 1, 1, 1])
    >>> b = np.array([1, 1, 1, 0, 0, 0, 0, 0, 0, 0])
    >>> modularity.match_cluster_labels(a, b)
    array([1, 1, 1, 0, 0, 0, 0, 2, 2, 2])

    If the source assignment has fewer clusters than the target the returned
    values may be discontinuous:

    >>> modularity.match_cluster_labels(b, a)
    array([0, 0, 0, 2, 2, 2, 2, 2, 2, 2])
    """
    # try and match the source to target
    src, tar = _get_relabels(source, target)

    # if there are a different number of clusters then handle elegantly.
    # elegantly here means we renumber the clusters so that they start at 1
    src_m = np.setdiff1d(np.unique(source), src)
    if len(src_m) > 0:
        tar_m = np.arange(tar.max() + 1, tar.max() + 1 + len(src_m))
        src, tar = np.append(src, src_m), np.append(tar, tar_m)

    # now re-label things based on the matched assignments
    sidx = src.argsort()
    src, tar = src[sidx], tar[sidx]
    matched = tar[np.searchsorted(src, source)]

    return matched


def match_assignments(assignments, target=None, seed=None):
    """
    Re-label clusters in columns of `assignments` to best match `target`.

    Uses :func:`~.cluster.match_cluster_labels` to align cluster assignments.

    Parameters
    ----------
    assignments : (N, M) array_like
        Array of `M` clustering assignments for `N` subjects
    target : (N,) array_like, optional
        Target clustering assignments to which all columns should be matched.
        If provided as an integer the relevant column in `assignments` will be
        selected. If not specified a (semi-)random column in `assignments` is
        chosen; because of the potential discontinuity introduced when matching
        an N-cluster solution to an N+1-cluster solution, the "random" target
        columns will be one `assignments` with the lowest cluster number. See
        Examples for more information. Default: None
    seed : {int, np.random.RandomState instance, None}, optional
        Seed for random number generation; only used if `target` is not
        provided. Default: None

    Returns
    -------
    assignments : (N, M) numpy.ndarray
        Provided array with re-labeled cluster solutions to better match across
        `M` assignments

    Examples
    --------
    >>> from netneurotools import modularity

    First we can construct a matrix of `N` samples clustered `M` times (in this
    case, `M` is three) . Since cluster labels are generally arbitrary we can
    see that, while the same clusters were found each time, they were given
    different labels:

    >>> assignments = np.array([[0, 0, 1],
    ...                         [0, 0, 1],
    ...                         [0, 0, 1],
    ...                         [1, 2, 0],
    ...                         [1, 2, 0],
    ...                         [1, 2, 0],
    ...                         [2, 1, 2],
    ...                         [2, 1, 2]])

    We would like to match the assignments so they're all the same. Since one
    of the columns will be randomly picked as the "target" solution, we provide
    a `seed` to ensure reproducibility in the selection:

    >>> modularity.match_assignments(assignments, seed=1234)
    array([[1, 1, 1],
           [1, 1, 1],
           [1, 1, 1],
           [0, 0, 0],
           [0, 0, 0],
           [0, 0, 0],
           [2, 2, 2],
           [2, 2, 2]])

    Alternatively, if `assignments` has clustering solutions with different
    numbers of clusters and no `target` is specified, the chosen `target` will
    be one of the columns with the smallest number of clusters:

    >>> assignments = np.array([[0, 0, 1],
    ...                         [0, 0, 1],
    ...                         [0, 0, 1],
    ...                         [1, 2, 0],
    ...                         [1, 2, 0],
    ...                         [1, 2, 0],
    ...                         [1, 1, 2],
    ...                         [1, 1, 2]])
    >>> modularity.match_assignments(assignments)
    array([[0, 0, 0],
           [0, 0, 0],
           [0, 0, 0],
           [1, 1, 1],
           [1, 1, 1],
           [1, 1, 1],
           [1, 2, 2],
           [1, 2, 2]])
    """
    assignments = np.asarray(assignments).copy()

    # pick a random assignment with the lowest # of clusters as "target"
    if target is None:
        rs = check_random_state(seed)
        cm = assignments.max(axis=0)
        mask = cm == cm.min()
        target = assignments[:, mask][:, rs.choice(mask.sum())]
    # use the specified column of the matrix
    elif isinstance(target, int):
        target = assignments[:, target]
    # assume that target is an iterable we can use (just check the length)
    else:
        if len(target) != len(assignments):
            raise ValueError('Length of target clustering solution must be '
                             'identical to length of provided array.')

    # iterate through all assignments and try and match them to the target
    for n, source in enumerate(assignments.T):
        assignments[:, n] = match_cluster_labels(source, target)

    return assignments


def reorder_assignments(assignments, consensus=None, col_sort=True,
                        row_sort=True, return_index=True, seed=None):
    """
    Relabel and reorders rows / columns of `assignments` to "look better".

    Relabels cluster solutions in `assignments` so that distinct clustering
    solutions have similar cluster labels. Then, swaps columns of `assignments`
    so that similar clustering solutions are placed near each other. Finally,
    swaps rows of `assignments` so that subjects with similar clustering
    profiles are placed near each other.

    Uses hierarchical clustering to generate re-ordering of columns and rows

    Parameters
    ----------
    assignments : (N, M) array_like
        Array of `M` clustering assignments for `N` subjects
    consensus : (N,) array_like, optional
        "Final" clustering solution for `N` subjects. If provided, reordering
        of rows will be constrained by cluster assignment. Default: None
    {row,col}_sort : bool, optional
        If True, sort the {rows, columns}. Default: True
    return_index : bool, optional
        Whether to return the row and column indices used to re-order
        `assignments` in addition to the re-ordered matrix. Default: True
    seed : {int, np.random.RandomState instance, None}, optional
        Seed for random number generation. Default: None

    Returns
    -------
    reordered : (N, M) numpy.ndarray
        Provided array with both rows and columns re-ordered
    index : tuple
        Indices used to reorder `assignments` to generate `reordered` output
    """

    def _reorder_rows(arr):
        """Return indices of rows in `arr` after hierarchical clustering."""
        link = hierarchy.linkage(arr, method='average', metric='hamming')
        return hierarchy.dendrogram(link, no_plot=True)['leaves']

    # first, relabel the columns to try and match across assignments; this will
    # make our reordering procedure work a bit better!
    assignments = match_assignments(assignments, seed=seed)

    if col_sort:
        # get max cluster number for each partition
        max_cl = assignments.max(axis=0)

        # if different assignments have different numbers of detected clusters
        if len(np.unique(max_cl)) > 1:
            # first sort based on the number of clusters in each assignment
            col_idx = max_cl.argsort()
            assignments, max_cl = assignments[:, col_idx], max_cl[col_idx]

            # then, within assignments with the same number of clusters reorder
            # assignments (columns)
            reordered, splits = [], np.where(np.diff(max_cl) != 0)[0] + 1
            col_idx = np.split(col_idx, splits)
            for n, cl in enumerate(np.split(assignments, splits, axis=1)):
                idx = _reorder_rows(cl.T)
                col_idx[n] = col_idx[n][idx]
                reordered += [cl[:, idx]]
            col_idx = list(np.hstack(col_idx))
            assignments = np.column_stack(reordered)

        # otherwise all assignments have same number of detected clusters so
        # just sort them all
        else:
            col_idx = list(_reorder_rows(assignments.T))
            assignments = assignments[:, col_idx]

    if row_sort:
        # if a consensus was provided reorder rows based on cluster assignment
        if consensus is not None:
            # sort subjects by their cluster assignment in the consensus for
            # each cluster, then reorder subjects (rows)
            reordered, row_idx = [], []
            for cl in np.unique(consensus):
                cl, = np.where(consensus == cl)
                idx = list(cl[_reorder_rows(assignments[cl])])
                reordered += [assignments[idx]]
                row_idx += idx
            assignments = np.vstack(reordered)

        # otherwise, just do a massive reordering of all the rows
        else:
            row_idx = list(_reorder_rows(assignments))
            assignments = assignments[row_idx]

    if return_index:
        return assignments, np.ix_(row_idx, col_idx)

    return assignments


def find_consensus(assignments, null_func=np.mean, return_agreement=False,
                   seed=None):
    """
    Find consensus clustering labels from cluster solutions in `assignments`.

    Parameters
    ----------
    assignments : (N, M) array_like
        Array of `M` clustering solutions for `N` samples (e.g., subjects,
        nodes, etc). Values of array should be integer-based cluster assignment
        labels
    null_func : callable, optional
        Function used to generate null model when performing consensus-based
        clustering. Must accept a 2D array as input and return a single value.
        Default: :func:`numpy.mean`
    return_agreement : bool, optional
        Whether to return the thresholded N x N agreement matrix used in
        generating the final consensus clustering solution. Default: False
    seed : {int, np.random.RandomState instance, None}, optional
        Seed for random number generation. Used when permuting cluster
        assignments during generation of null model. Default: None

    Returns
    -------
    consensus : (N,) numpy.ndarray
        Consensus cluster labels

    References
    ----------
    Bassett, D. S., Porter, M. A., Wymbs, N. F., Grafton, S. T., Carlson,
    J. M., & Mucha, P. J. (2013). Robust detection of dynamic community
    structure in networks. Chaos: An Interdisciplinary Journal of Nonlinear
    Science, 23(1), 013142.
    """
    rs = check_random_state(seed)
    samp, comm = assignments.shape

    # create agreement matrix from input community assignments and convert to
    # probability matrix by dividing by `comm`
    agreement = bct.clustering.agreement(assignments, buffsz=samp) / comm

    # generate null agreement matrix and use to create threshold
    null_assign = np.column_stack([rs.permutation(i) for i in assignments.T])
    null_agree = bct.clustering.agreement(null_assign, buffsz=samp) / comm
    threshold = null_func(null_agree)

    # run consensus clustering on agreement matrix after thresholding
    consensus = bct.clustering.consensus_und(agreement, threshold, 10)

    if return_agreement:
        return consensus.astype(int), agreement * (agreement > threshold)

    return consensus.astype(int)


def consensus_modularity(adjacency, gamma=1, B='modularity',
                         repeats=250, null_func=np.mean, seed=None):
    """
    Find community assignments from `adjacency` through consensus.

    Performs `repeats` iterations of community detection on `adjacency` and
    then uses consensus clustering on the resulting community assignments.

    Parameters
    ----------
    adjacency : (N, N) array_like
        Adjacency matrix (weighted/non-weighted) on which to perform consensus
        community detection.
    gamma : float, optional
        Resolution parameter for modularity maximization. Default: 1
    B : str or (N, N) array_like, optional
        Null model to use for consensus clustering. If `str`, must be one of
        ['modularity', 'potts', 'negative_sym', 'negative_asym']. Default:
        'modularity'
    repeats : int, optional
        Number of times to repeat Louvain algorithm clustering. Default: 250
    null_func : callable, optional
        Function used to generate null model when performing consensus-based
        clustering. Must accept a 2D array as input and return a single value.
        Default: `np.mean`
    seed : {int, np.random.RandomState instance, None}, optional
        Seed for random number generation. Default: None

    Returns
    -------
    consensus : (N,) np.ndarray
        Consensus-derived community assignments
    Q_all : array_like
        Optimized modularity over all `repeats` community assignments
    zrand_all : array_like
        z-Rand score over all pairs of `repeats` community assignment vectors

    References
    ----------
    Bassett, D. S., Porter, M. A., Wymbs, N. F., Grafton, S. T., Carlson,
    J. M., & Mucha, P. J. (2013). Robust detection of dynamic community
    structure in networks. Chaos: An Interdisciplinary Journal of Nonlinear
    Science, 23(1), 013142.
    """
    # generate community partitions `repeat` times
    comms, Q_all = zip(*[bct.community_louvain(adjacency, gamma=gamma, B=B)
                         for i in range(repeats)])
    comms = np.column_stack(comms)

    # find consensus cluster assignments across all partitoning solutions
    consensus = find_consensus(comms, null_func=null_func, seed=seed)

    # get z-rand statistics for partition similarity (n.b. can take a while)
    zrand_all = _zrand_partitions(comms)

    return consensus, np.array(Q_all), zrand_all


def _dummyvar(labels):
    """
    Generate dummy-coded array from provided community assignment `labels`.

    Parameters
    ----------
    labels : (N,) array_like
        Labels assigning `N` samples to `G` groups

    Returns
    -------
    ci : (N, G) numpy.ndarray
        Dummy-coded array where 1 indicates that a sample belongs to a group
    """
    comms = np.unique(labels)

    ci = np.zeros((len(labels), len(comms)))
    for n, grp in enumerate(comms):
        ci[:, n] = labels == grp

    return ci


def zrand(X, Y):
    """
    Calculate the z-Rand index of two community assignments.

    Parameters
    ----------
    X, Y : (n, 1) array_like
        Community assignment vectors to compare

    Returns
    -------
    z_rand : float
        Z-rand index

    References
    ----------
    Amanda L. Traud, Eric D. Kelsic, Peter J. Mucha, and Mason A. Porter.
    (2011). Comparing Community Structure to Characteristics in Online
    Collegiate Social Networks. SIAM Review, 53, 526-543.
    """
    if X.ndim > 1 or Y.ndim > 1:
        if X.shape[-1] > 1 or Y.shape[-1] > 1:
            raise ValueError('X and Y must have only one-dimension each. '
                             'Please check inputs.')

    Xf = X.flatten()
    Yf = Y.flatten()

    n = len(Xf)
    indx, indy = _dummyvar(Xf), _dummyvar(Yf)
    Xa = indx.dot(indx.T)
    Ya = indy.dot(indy.T)

    M = n * (n - 1) / 2
    M1 = Xa.nonzero()[0].size / 2
    M2 = Ya.nonzero()[0].size / 2

    wab = np.logical_and(Xa, Ya).nonzero()[0].size / 2

    mod = n * (n**2 - 3 * n - 2)
    C1 = mod - (8 * (n + 1) * M1) + (4 * np.power(indx.sum(0), 3).sum())
    C2 = mod - (8 * (n + 1) * M2) + (4 * np.power(indy.sum(0), 3).sum())

    a = M / 16
    b = ((4 * M1 - 2 * M)**2) * ((4 * M2 - 2 * M)**2) / (256 * (M**2))
    c = C1 * C2 / (16 * n * (n - 1) * (n - 2))
    d = ((((4 * M1 - 2 * M)**2) - (4 * C1) - (4 * M))
         * (((4 * M2 - 2 * M)**2) - (4 * C2) - (4 * M))
         / (64 * n * (n - 1) * (n - 2) * (n - 3)))

    sigw2 = a - b + c + d
    # catch any negatives
    if sigw2 < 0:
        return 0
    z_rand = (wab - ((M1 * M2) / M)) / np.sqrt(sigw2)

    return z_rand


def _zrand_partitions(communities):
    """
    Calculate z-Rand for all pairs of assignments in `communities`.

    Iterates through every pair of community assignment vectors in
    `communities` and calculates the z-Rand score to assess their similarity.

    Parameters
    ----------
    communities : (S, R) array_like
        Community assignments for `S` samples over `R` partitions

    Returns
    -------
    all_zrand : array_like
        z-Rand score over all pairs of `R` partitions of community assignments
    """
    n_partitions = communities.shape[-1]
    all_zrand = np.zeros(int(n_partitions * (n_partitions - 1) / 2))

    for c1 in range(n_partitions):
        for c2 in range(c1 + 1, n_partitions):
            idx = int((c1 * n_partitions) + c2 - ((c1 + 1) * (c1 + 2) // 2))
            all_zrand[idx] = zrand(communities[:, c1], communities[:, c2])

    return all_zrand


if has_numba:
    _dummyvar = njit(_dummyvar)
    zrand = njit(zrand)
    _zrand_partitions = njit(_zrand_partitions, parallel=True)


def get_modularity(adjacency, comm, gamma=1):
    """
    Calculate modularity contribution for each community in `comm`.

    Parameters
    ----------
    adjacency : (N, N) array_like
        Adjacency (e.g., correlation) matrix
    comm : (N,) array_like
        Community assignment vector splitting `N` subjects into `G` groups
    gamma : float, optional
        Resolution parameter used in original modularity maximization.
        Default: 1

    Returns
    -------
    comm_q : (G,) ndarray
        Relative modularity for each community

    See Also
    --------
    netneurotools.modularity.get_modularity_z
    netneurotools.modularity.get_modularity_sig
    """
    adjacency, comm = np.asarray(adjacency), np.asarray(comm)
    s = adjacency.sum()
    B = adjacency - (gamma * np.outer(adjacency.sum(axis=1),
                                      adjacency.sum(axis=0)) / s)

    # find modularity contribution of each community
    communities = np.unique(comm)
    comm_q = np.empty(shape=communities.size)
    for n, ci in enumerate(communities):
        inds = comm == ci
        comm_q[n] = B[np.ix_(inds, inds)].sum() / s

    return comm_q


def get_modularity_z(adjacency, comm, gamma=1, n_perm=10000, seed=None):
    """
    Calculate average z-score of community assignments by permutation.

    Parameters
    ----------
    adjacency : (N, N) array_like
        Adjacency (correlation) matrix
    comm : (N,) array_like
        Community assignment vector splitting `N` subjects into `G` groups
    gamma : float, optional
        Resolution parameter used in original modularity maximization.
        Default: 1
    n_perm : int, optional
        Number of permutations. Default: 10000
    seed : {int, np.random.RandomState instance, None}, optional
        Seed for random number generation. Default: None

    Returns
    -------
    q_z : float
        Average Z-score of modularity of communities

    See Also
    --------
    netneurotools.modularity.get_modularity
    netneurotools.modularity.get_modularity_sig
    """
    rs = check_random_state(seed)

    real_qs = get_modularity(adjacency, comm, gamma)
    simu_qs = np.empty(shape=(np.unique(comm).size, n_perm))
    for perm in range(n_perm):
        simu_qs[:, perm] = get_modularity(adjacency,
                                          rs.permutation(comm),
                                          gamma)

    # avoid instances where dist.std(1) == 0
    std = simu_qs.std(axis=1)
    if std == 0:
        return np.mean(real_qs - simu_qs.mean(axis=1))
    else:
        return np.mean((real_qs - simu_qs.mean(axis=1)) / std)


def get_modularity_sig(adjacency, comm, gamma=1, n_perm=10000, alpha=0.01,
                       seed=None):
    """
    Calculate significance of community assignments in `comm` by permutation.

    Parameters
    ----------
    adjacency : (N, N) array_like
        Adjacency (correlation) matrix
    comm : (N,) array_like
        Community assignment vector
    gamma : float
        Resolution parameter used in original modularity maximization
    n_perm : int, optional
        Number of permutations to test against. Default: 10000
    alpha : (0,1) float, optional
        Alpha level to assess significance. Default: 0.01
    seed : {int, np.random.RandomState instance, None}, optional
        Seed for random number generation. Default: None

    Returns
    -------
    ndarray
        Significance of each community in `comm` (boolean)

    See Also
    --------
    netneurotools.modularity.get_modularity_z
    netneurotools.modularity.get_modularity_sig
    """
    rs = check_random_state(seed)

    real_qs = get_modularity(adjacency, comm, gamma)
    simu_qs = np.empty(shape=(np.unique(comm).size, n_perm))
    for perm in range(n_perm):
        simu_qs[:, perm] = get_modularity(adjacency,
                                          rs.permutation(comm),
                                          gamma)

    q_sig = real_qs > np.percentile(simu_qs, 100 * (1 - alpha), axis=1)

    return q_sig



"""Functions for handling datasets."""


from .fetch_template import (
    fetch_fsaverage, fetch_fsaverage_curated,
    fetch_hcp_standards, fetch_fslr_curated,
    fetch_civet, fetch_civet_curated,
    fetch_conte69, fetch_yerkes19
)


from .fetch_atlas import (
    # cortical
    fetch_cammoun2012, fetch_schaefer2018, fetch_mmpall,
    # subcortical
    fetch_pauli2018, fetch_tian2020msa,
    # annotation
    fetch_voneconomo
)


from .fetch_project import (
    # old projects
    fetch_vazquez_rodriguez2019, fetch_mirchi2018,
    # new projects
    fetch_hansen_manynetworks, fetch_hansen_receptors,
    fetch_hansen_genescognition, fetch_hansen_brainstemfc,
    fetch_shafiei_megfmrimapping, fetch_shafiei_megdynamics,
    fetch_suarez_mami,
    # example data
    fetch_famous_gmat,
    # resources
    fetch_neurosynth
)

from .datasets_utils import (
    FREESURFER_IGNORE, _get_freesurfer_subjid
)


__all__ = [
    # fetch_template
    'fetch_fsaverage', 'fetch_fsaverage_curated',
    'fetch_hcp_standards', 'fetch_fslr_curated',
    'fetch_civet', 'fetch_civet_curated',
    'fetch_conte69', 'fetch_yerkes19',
    # fetch_atlas
    'fetch_cammoun2012', 'fetch_schaefer2018', 'fetch_mmpall',
    'fetch_pauli2018', 'fetch_tian2020msa',
    'fetch_voneconomo',
    # fetch_project
    'fetch_vazquez_rodriguez2019', 'fetch_mirchi2018',
    'fetch_hansen_manynetworks', 'fetch_hansen_receptors',
    'fetch_hansen_genescognition', 'fetch_hansen_brainstemfc',
    'fetch_shafiei_megfmrimapping', 'fetch_shafiei_megdynamics',
    'fetch_suarez_mami',
    'fetch_famous_gmat',
    'fetch_neurosynth',
    # datasets_utils
    'FREESURFER_IGNORE', '_get_freesurfer_subjid'
]



"""Functions for fetching project data."""

import os
import numpy as np

from sklearn.utils import Bunch

from .datasets_utils import _get_data_dir, _get_reference_info, fetch_file

from ._mirchi2018 import _get_fc, _get_panas


def fetch_vazquez_rodriguez2019(force=False, data_dir=None, verbose=1):
    """
    Download files from Vazquez-Rodriguez et al., 2019, PNAS.

    This dataset contains one file: rsquared_gradient.csv, which contains
    two columns: rsquared and gradient.

    If you used this data, please cite [1]_.

    Returns
    -------
    data : :class:`sklearn.utils.Bunch`
        Dictionary-like object with fetched data.

    Other Parameters
    ----------------
    force : bool, optional
        If True, will overwrite existing dataset. Default: False
    data_dir : str, optional
        Path to use as data directory. If not specified, will check for
        environmental variable 'NNT_DATA'; if that is not set, will use
        `~/nnt-data` instead. Default: None
    verbose : int, optional
        Modifies verbosity of download, where higher numbers mean more updates.
        Default: 1

    References
    ----------
    .. [1] Bertha V\u00e1zquez-Rodr\u00edguez, Laura E Su\u00e1rez, Ross D
        Markello, Golia Shafiei, Casey Paquola, Patric Hagmann, Martijn P Van
        Den Heuvel, Boris C Bernhardt, R Nathan Spreng, and Bratislav Misic.
        Gradients of structure\u2013function tethering across neocortex.
        Proceedings of the National Academy of Sciences,
        116(42):21219\u201321227, 2019.
    """
    dataset_name = "ds-vazquez_rodriguez2019"
    _get_reference_info(dataset_name, verbose=verbose)

    fetched = fetch_file(dataset_name, force=force, data_dir=data_dir, verbose=verbose)

    # load data
    rsq, grad = np.loadtxt(
        fetched / "rsquared_gradient.csv", delimiter=",", skiprows=1
    ).T
    data = {"rsquared": rsq, "gradient": grad}

    return Bunch(**data)


def fetch_mirchi2018(force=False, data_dir=None, verbose=1):
    """
    Download (and creates) dataset for replicating Mirchi et al., 2018, SCAN.

    Parameters
    ----------
    data_dir : str, optional
        Directory to check for existing data files (if they exist) or to save
        generated data files. Files should be named mirchi2018_fc.npy and
        mirchi2018_panas.csv for the functional connectivity and behavioral
        data, respectively.

    Returns
    -------
    X : (73, 198135) numpy.ndarray
        Functional connections from MyConnectome rsfMRI time series data
    Y : (73, 13) numpy.ndarray
        PANAS subscales from MyConnectome behavioral data
    """
    data_dir = _get_data_dir(data_dir=data_dir) / "ds-mirchi2018"
    data_dir.mkdir(exist_ok=True, parents=True)

    X_fname = data_dir / "myconnectome_fc.npy"
    Y_fname = data_dir / "myconnectome_panas.csv"

    if not os.path.exists(X_fname):
        X = _get_fc(verbose=verbose)
        np.save(X_fname, X, allow_pickle=False)
    else:
        X = np.load(X_fname, allow_pickle=False)

    if not os.path.exists(Y_fname):
        Y = _get_panas()
        np.savetxt(
            Y_fname,
            np.column_stack(list(Y.values())),
            header=",".join(Y.keys()),
            delimiter=",",
            fmt="%i",
        )
        # convert dictionary to structured array before returning
        Y = np.array(
            [tuple(row) for row in np.column_stack(list(Y.values()))],
            dtype=dict(names=list(Y.keys()), formats=["i8"] * len(Y)),
        )
    else:
        Y = np.genfromtxt(Y_fname, delimiter=",", names=True, dtype=int)

    return X, Y


def fetch_hansen_manynetworks(force=False, data_dir=None, verbose=1):
    """
    Download files from Hansen et al., 2023, PLOS Biology.

    This dataset contains

    If you used this data, please cite [1]_.

    Returns
    -------
    filenames : :class:`sklearn.utils.Bunch`
        Dictionary-like object with fetched data.

    Other Parameters
    ----------------
    force : bool, optional
        If True, will overwrite existing dataset. Default: False
    data_dir : str, optional
        Path to use as data directory. If not specified, will check for
        environmental variable 'NNT_DATA'; if that is not set, will use
        `~/nnt-data` instead. Default: None
    verbose : int, optional
        Modifies verbosity of download, where higher numbers mean more updates.
        Default: 1

    References
    ----------
    .. [1]
    """
    dataset_name = "ds-hansen_manynetworks"
    _get_reference_info(dataset_name, verbose=verbose)

    fetched = fetch_file(dataset_name, force=force, data_dir=data_dir, verbose=verbose)

    # _parc = {
    #     "cammoun033": "Cammoun033",
    #     "schaefer100": "Schaefer100",
    #     "schaefer400": "Schaefer400",
    # }

    # data = {
    #     parc: {
    #         "bigbrain": fetched
    #         / f"data/{_parc[parc]}/bigbrain_intensities.csv",
    #         "cognitive": fetched
    #         / f"data/{_parc[parc]}/cognitive_similarity.npy",
    #         "sc": fetched / f"data/{_parc[parc]}/consensusSC.npy",
    #         "sc_wei": fetched / f"data/{_parc[parc]}/consensusSC_wei.npy",
    #         "ephys": fetched
    #         / f"data/{_parc[parc]}/electrophysiological_connectivity.npy",
    #         "gene": fetched / f"data/{_parc[parc]}/gene_coexpression.npy",
    #         "megfc": fetched
    #         / f"data/{_parc[parc]}/groupFCmeg_aec_orth_{_parc[parc]}.npy.npz",
    #         "haemodynamic": fetched
    #         / f"data/{_parc[parc]}/haemodynamic_connectivity.npy",
    #         "laminar": fetched / f"data/{_parc[parc]}/laminar_similarity.npy",
    #         "metabolic": fetched
    #         / f"data/{_parc[parc]}/metabolic_connectivity.npy",
    #         "receptor": fetched
    #         / f"data/{_parc[parc]}/receptor_similarity.npy",
    #         "temporal": fetched
    #         / f"data/{_parc[parc]}/temporal_similarity.npy",
    #         "voneconomo": fetched
    #         / f"data/{_parc[parc]}/voneconomo_{_parc[parc]}.csv",
    #     }
    #     for parc in ["cammoun033", "schaefer100", "schaefer400"]
    # }

    return fetched


def fetch_hansen_receptors(force=False, data_dir=None, verbose=1):
    """
    Download files from Hansen et al., 2022, Nature Neuroscience.

    This dataset contains

    If you used this data, please cite [1]_.

    .. important::
       If you use this dataset, please also cite the original data sources.
       See the original publication or `neuromaps <https://netneurolab.github.io/neuromaps/listofmaps.html>`_
       for more information.

    Returns
    -------
    filenames : :class:`sklearn.utils.Bunch`
        Dictionary-like object with fetched data.

    Other Parameters
    ----------------
    force : bool, optional
        If True, will overwrite existing dataset. Default: False
    data_dir : str, optional
        Path to use as data directory. If not specified, will check for
        environmental variable 'NNT_DATA'; if that is not set, will use
        `~/nnt-data` instead. Default: None
    verbose : int, optional
        Modifies verbosity of download, where higher numbers mean more updates.
        Default: 1

    References
    ----------
    .. [1] Justine Y Hansen, Golia Shafiei, Ross D Markello, Kelly Smart, Sylvia
        ML Cox, Martin N\u00f8rgaard, Vincent Beliveau, Yanjun Wu,
        Jean-Dominique Gallezot, \u00c9tienne Aumont, and others. Mapping
        neurotransmitter systems to the structural and functional organization
        of the human neocortex. Nature neuroscience, 25(11):1569\u20131581,
        2022.
    """
    dataset_name = "ds-hansen_receptors"
    _get_reference_info(dataset_name, verbose=verbose)

    fetched = fetch_file(dataset_name, force=force, data_dir=data_dir, verbose=verbose)

    return fetched


def fetch_hansen_genescognition(force=False, data_dir=None, verbose=1):
    """Download files from Hansen et al., 2021, Nature Human Behaviour.

    This dataset contains

    If you used this data, please cite [1]_.

    Returns
    -------
    filenames : :class:`sklearn.utils.Bunch`
        Dictionary-like object with fetched data.

    Other Parameters
    ----------------
    force : bool, optional
        If True, will overwrite existing dataset. Default: False
    data_dir : str, optional
        Path to use as data directory. If not specified, will check for
        environmental variable 'NNT_DATA'; if that is not set, will use
        `~/nnt-data` instead. Default: None
    verbose : int, optional
        Modifies verbosity of download, where higher numbers mean more updates.
        Default: 1

    References
    ----------
    .. [1] Justine Y Hansen, Ross D Markello, Jacob W Vogel, Jakob Seidlitz,
        Danilo Bzdok, and Bratislav Misic. Mapping gene transcription and
        neurocognition across human neocortex. Nature Human Behaviour,
        5(9):1240\u20131250, 2021.
    """
    dataset_name = "ds-hansen_genescognition"
    _get_reference_info(dataset_name, verbose=verbose)

    fetched = fetch_file(dataset_name, force=force, data_dir=data_dir, verbose=verbose)

    return fetched


def fetch_hansen_brainstemfc(force=False, data_dir=None, verbose=1):
    """Download files from Hansen et al., 2024.

    This dataset contains

    If you used this data, please cite [1]_.

    Returns
    -------
    filenames : :class:`sklearn.utils.Bunch`
        Dictionary-like object with fetched data.

    Other Parameters
    ----------------
    force : bool, optional
        If True, will overwrite existing dataset. Default: False
    data_dir : str, optional
        Path to use as data directory. If not specified, will check for
        environmental variable 'NNT_DATA'; if that is not set, will use
        `~/nnt-data` instead. Default: None
    verbose : int, optional
        Modifies verbosity of download, where higher numbers mean more updates.
        Default: 1

    References
    ----------
    .. [1] Justine Y Hansen, Simone Cauzzo, Kavita Singh, Mar\u00eda Guadalupe
        Garc\u00eda-Gomar, James M Shine, Marta Bianciardi, and Bratislav Misic.
        Integrating brainstem and cortical functional architectures. Nature
        Neuroscience, pages 1\u201312, 2024.
    """
    dataset_name = "ds-hansen_brainstemfc"
    _get_reference_info(dataset_name, verbose=verbose)

    fetched = fetch_file(dataset_name, force=force, data_dir=data_dir, verbose=verbose)

    return fetched


def fetch_shafiei_megfmrimapping(force=False, data_dir=None, verbose=1):
    """Download files from Shafiei et al., 2022.

    This dataset contains

    If you used this data, please cite [1]_.

    Returns
    -------
    filenames : :class:`sklearn.utils.Bunch`
        Dictionary-like object with fetched data.

    Other Parameters
    ----------------
    force : bool, optional
        If True, will overwrite existing dataset. Default: False
    data_dir : str, optional
        Path to use as data directory. If not specified, will check for
        environmental variable 'NNT_DATA'; if that is not set, will use
        `~/nnt-data` instead. Default: None
    verbose : int, optional
        Modifies verbosity of download, where higher numbers mean more updates.
        Default: 1

    References
    ----------
    .. [1] Golia Shafiei, Sylvain Baillet, and Bratislav Misic. Human
        electromagnetic and haemodynamic networks systematically converge in
        unimodal cortex and diverge in transmodal cortex. PLoS biology,
        20(8):e3001735, 2022.
    """
    dataset_name = "ds-shafiei_megfmrimapping"
    _get_reference_info(dataset_name, verbose=verbose)

    fetched = fetch_file(dataset_name, force=force, data_dir=data_dir, verbose=verbose)

    return fetched


def fetch_shafiei_megdynamics(force=False, data_dir=None, verbose=1):
    """Download files from Shafiei et al., 2023.

    This dataset contains

    If you used this data, please cite [1]_.

    Returns
    -------
    filenames : :class:`sklearn.utils.Bunch`
        Dictionary-like object with fetched data.

    Other Parameters
    ----------------
    force : bool, optional
        If True, will overwrite existing dataset. Default: False
    data_dir : str, optional
        Path to use as data directory. If not specified, will check for
        environmental variable 'NNT_DATA'; if that is not set, will use
        `~/nnt-data` instead. Default: None
    verbose : int, optional
        Modifies verbosity of download, where higher numbers mean more updates.
        Default: 1

    References
    ----------
    .. [1] Golia Shafiei, Ben D Fulcher, Bradley Voytek, Theodore D
        Satterthwaite, Sylvain Baillet, and Bratislav Misic. Neurophysiological
        signatures of cortical micro-architecture. Nature communications,
        14(1):6000, 2023.
    """
    dataset_name = "ds-shafiei_megdynamics"
    _get_reference_info(dataset_name, verbose=verbose)

    fetched = fetch_file(dataset_name, force=force, data_dir=data_dir, verbose=verbose)

    return fetched


def fetch_suarez_mami(force=False, data_dir=None, verbose=1):
    """Download files from Suarez et al., 2022, eLife.

    This dataset contains

    If you used this data, please cite [1]_ and [2]_.

    Returns
    -------
    filenames : :class:`sklearn.utils.Bunch`
        Dictionary-like object with fetched data.

    Other Parameters
    ----------------
    force : bool, optional
        If True, will overwrite existing dataset. Default: False
    data_dir : str, optional
        Path to use as data directory. If not specified, will check for
        environmental variable 'NNT_DATA'; if that is not set, will use
        `~/nnt-data` instead. Default: None
    verbose : int, optional
        Modifies verbosity of download, where higher numbers mean more updates.
        Default: 1

    References
    ----------
    .. [1] Laura E Suarez, Yossi Yovel, Martijn P van den Heuvel, Olaf Sporns,
        Yaniv Assaf, Guillaume Lajoie, and Bratislav Misic. A connectomics-based
        taxonomy of mammals. Elife, 11:e78635, 2022.
    .. [2] Yaniv Assaf, Arieli Bouznach, Omri Zomet, Assaf Marom, and Yossi
        Yovel. Conservation of brain connectivity and wiring across the
        mammalian class. Nature Neuroscience, 23(7):805\u2013808, 2020.
    """
    dataset_name = "ds-suarez_mami"
    _get_reference_info(dataset_name, verbose=verbose)

    fetched = fetch_file(dataset_name, force=force, data_dir=data_dir, verbose=verbose)

    return fetched


def fetch_famous_gmat(dataset, force=False, data_dir=None, verbose=1):
    """
    Download files from multi-species connectomes.

    This dataset contains

    If you used this data, please cite celegans [1]_, drosophila [2]_, human
    [3]_, macaque_markov [4]_, macaque_modha [5]_, mouse [6]_, rat [7]_.

    Parameters
    ----------
    dataset : str
        Specifies which dataset to download.

    Returns
    -------
    data : :class:`sklearn.utils.Bunch`
        Dictionary-like object with, at a minimum, keys ['conn', 'labels',
        'ref'] providing connectivity / correlation matrix, region labels, and
        relevant reference. Other possible keys include 'dist' (an array of
        Euclidean distances between regions of 'conn'), 'coords' (an array of
        xyz coordinates for regions of 'conn'), 'acronyms' (an array of
        acronyms for regions of 'conn'), and 'networks' (an array of network
        affiliations for regions of 'conn').

    Other Parameters
    ----------------
    force : bool, optional
        If True, will overwrite existing dataset. Default: False
    data_dir : str, optional
        Path to use as data directory. If not specified, will check for
        environmental variable 'NNT_DATA'; if that is not set, will use
        `~/nnt-data` instead. Default: None
    verbose : int, optional
        Modifies verbosity of download, where higher numbers mean more updates.
        Default: 1

    References
    ----------
    .. [1] Lav R Varshney, Beth L Chen, Eric Paniagua, David H Hall, and Dmitri
        B Chklovskii. Structural properties of the caenorhabditis elegans
        neuronal network. PLoS computational biology, 7(2):e1001066, 2011.
    .. [2] Ann-Shyn Chiang, Chih-Yung Lin, Chao-Chun Chuang, Hsiu-Ming Chang,
        Chang-Huain Hsieh, Chang-Wei Yeh, Chi-Tin Shih, Jian-Jheng Wu, Guo-Tzau
        Wang, Yung-Chang Chen, and others. Three-dimensional reconstruction of
        brain-wide wiring networks in drosophila at single-cell resolution.
        Current biology, 21(1):1\u201311, 2011.
    .. [3] Alessandra Griffa, Yasser Alem\u00e1n-G\u00f3mez, and Patric Hagmann.
        Structural and functional connectome from 70 young healthy adults [data
        set]. Zenodo, 2019.
    .. [4] Nikola T Markov, Maria Ercsey-Ravasz, Camille Lamy, Ana Rita Ribeiro
        Gomes, Lo\u00efc Magrou, Pierre Misery, Pascale Giroud, Pascal Barone,
        Colette Dehay, Zolt\u00e1n Toroczkai, and others. The role of long-range
        connections on the specificity of the macaque interareal cortical
        network. Proceedings of the National Academy of Sciences,
        110(13):5187\u20135192, 2013.
    .. [5] Dharmendra S Modha and Raghavendra Singh. Network architecture of the
        long-distance pathways in the macaque brain. Proceedings of the National
        Academy of Sciences, 107(30):13485\u201313490, 2010.
    .. [6] Mikail Rubinov, Rolf JF Ypma, Charles Watson, and Edward T Bullmore.
        Wiring cost and topological participation of the mouse brain connectome.
        Proceedings of the National Academy of Sciences,
        112(32):10032\u201310037, 2015.
    .. [7] Mihail Bota, Olaf Sporns, and Larry W Swanson. Architecture of the
        cerebral cortical association connectome underlying cognition.
        Proceedings of the National Academy of Sciences,
        112(16):E2093\u2013E2101, 2015.
    """
    available_connectomes = [
        "celegans",
        "drosophila",
        "human_func_scale033",
        "human_func_scale060",
        "human_func_scale125",
        "human_func_scale250",
        "human_func_scale500",
        "human_struct_scale033",
        "human_struct_scale060",
        "human_struct_scale125",
        "human_struct_scale250",
        "human_struct_scale500",
        "macaque_markov",
        "macaque_modha",
        "mouse",
        "rat",
    ]

    if dataset not in available_connectomes:
        raise ValueError(
            f"Provided dataset {dataset} not available; "
            f"must be one of {available_connectomes}"
        )

    base_dataset_name = "ds-famous_gmat"
    _get_reference_info(base_dataset_name, verbose=verbose)

    fetched = fetch_file(
        base_dataset_name, force=force, data_dir=data_dir, verbose=verbose
    )

    data = {}
    for f in (fetched / dataset).glob("*.csv"):
        try:
            data[f.stem] = np.loadtxt(f, delimiter=",")
        except ValueError:
            data[f.stem] = np.loadtxt(f, delimiter=",", dtype=str)

    return Bunch(**data)


def fetch_neurosynth():
    """Download Neurosynth data."""
    pass



"""Functions for fetching atlas data."""

import itertools


from sklearn.utils import Bunch

from .datasets_utils import SURFACE, _get_reference_info, fetch_file


def fetch_cammoun2012(
    version="MNI152NLin2009aSym", force=False, data_dir=None, verbose=1
):
    """
    Download files for Cammoun et al., 2012 multiscale parcellation.

    This dataset contains

    If you used this data, please cite 1_.

    Parameters
    ----------
    version : str, optional
        Specifies which version of the dataset to download, where
        'MNI152NLin2009aSym' will return .nii.gz atlas files defined in MNI152
        space, 'fsaverageX' will return .annot files defined in fsaverageX
        space (FreeSurfer 6.0.1), 'fslr32k' will return .label.gii files in
        fs_LR_32k HCP space, and 'gcs' will return FreeSurfer-style .gcs
        probabilistic atlas files for generating new, subject-specific
        parcellations. Default: 'MNI152NLin2009aSym'

    Returns
    -------
    filenames : :class:`sklearn.utils.Bunch`
        Dictionary-like object with keys ['scale033', 'scale060', 'scale125',
        'scale250', 'scale500'], where corresponding values are lists of
        filepaths to downloaded parcellation files.

    Other Parameters
    ----------------
    force : bool, optional
        If True, will overwrite existing dataset. Default: False
    data_dir : str, optional
        Path to use as data directory. If not specified, will check for
        environmental variable 'NNT_DATA'; if that is not set, will use
        `~/nnt-data` instead. Default: None
    verbose : int, optional
        Modifies verbosity of download, where higher numbers mean more updates.
        Default: 1

    Notes
    -----
    License: https://raw.githubusercontent.com/LTS5/cmp/master/COPYRIGHT

    References
    ----------
    .. [1] Leila Cammoun, Xavier Gigandet, Djalel Meskaldji, Jean Philippe
        Thiran, Olaf Sporns, Kim Q Do, Philippe Maeder, Reto Meuli, and Patric
        Hagmann. Mapping the human connectome at multiple scales with diffusion
        spectrum mri. Journal of neuroscience methods, 203(2):386\u2013397,
        2012.
    """
    versions = [
        "gcs",
        "fsaverage",
        "fsaverage5",
        "fsaverage6",
        "fslr32k",
        "MNI152NLin2009aSym",
    ]
    if version not in versions:
        raise ValueError(
            f"The version of Cammoun et al., 2012 parcellation "
            f"requested {version} does not exist. Must be one of {versions}"
        )

    dataset_name = "atl-cammoun2012"
    _get_reference_info(dataset_name, verbose=verbose)

    keys = ["scale033", "scale060", "scale125", "scale250", "scale500"]

    fetched = fetch_file(
        dataset_name, keys=version, force=force, data_dir=data_dir, verbose=verbose
    )

    if version == "MNI152NLin2009aSym":
        _fname = "atl-Cammoun2012_space-MNI152NLin2009aSym_res-{}_deterministic.nii.gz"
        data = {
            k: fetched
            / _fname.format(k[-3:])
            for k in keys
        }
        data["info"] = fetched / "atl-Cammoun2012_space-MNI152NLin2009aSym_info.csv"
    elif version == "fslr32k":
        _fname = "atl-Cammoun2012_space-fslr32k_res-{}_hemi-{}_deterministic.label.gii"
        data = {
            k: SURFACE(
                fetched / _fname.format(k[-3:], "L"),
                fetched / _fname.format(k[-3:], "R")
            )
            for k in keys
        }
    elif version in ("fsaverage", "fsaverage5", "fsaverage6"):
        _fname = "atl-Cammoun2012_space-{}_res-{}_hemi-{}_deterministic.annot"
        data = {
            k: SURFACE(
                fetched / _fname.format(version, k[-3:], "L"),
                fetched / _fname.format(version, k[-3:], "R")
            )
            for k in keys
        }
    else:
        data = {
            k: [
                fetched / f"atl-Cammoun2012_res-{k[5:]}_hemi-L_probabilistic.gcs",
                fetched / f"atl-Cammoun2012_res-{k[5:]}_hemi-R_probabilistic.gcs",
            ]
            for k in keys[:-1]
        }
        data[keys[-1]] = list(
            itertools.chain.from_iterable(
                [
                    [
                        fetched
                        / f"atl-Cammoun2012_res-{k[5:]}_hemi-L_probabilistic.gcs",
                        fetched
                        / f"atl-Cammoun2012_res-{k[5:]}_hemi-R_probabilistic.gcs",
                    ]
                    for k in ["scale500v1", "scale500v2", "scale500v3"]
                ]
            )
        )

    return Bunch(**data)


def fetch_schaefer2018(version="fsaverage", force=False, data_dir=None, verbose=1):
    """
    Download FreeSurfer .annot files for Schaefer et al., 2018 parcellation.

    This dataset contains

    If you used this data, please cite 1_.

    Parameters
    ----------
    version : {'fsaverage', 'fsaverage5', 'fsaverage6', 'fslr32k'}
        Specifies which surface annotation files should be matched to. Default:
        'fsaverage'

    Returns
    -------
    filenames : :class:`sklearn.utils.Bunch`
        Dictionary-like object with keys of format '{}Parcels{}Networks' where
        corresponding values are the left/right hemisphere annotation files

    Other Parameters
    ----------------
    force : bool, optional
        If True, will overwrite existing dataset. Default: False
    data_dir : str, optional
        Path to use as data directory. If not specified, will check for
        environmental variable 'NNT_DATA'; if that is not set, will use
        `~/nnt-data` instead. Default: None
    verbose : int, optional
        Modifies verbosity of download, where higher numbers mean more updates.
        Default: 1

    Notes
    -----
    License: https://github.com/ThomasYeoLab/CBIG/blob/master/LICENSE.md

    References
    ----------
    .. [1] Alexander Schaefer, Ru Kong, Evan M Gordon, Timothy O Laumann,
        Xi-Nian Zuo, Avram J Holmes, Simon B Eickhoff, and BT Thomas Yeo.
        Local-global parcellation of the human cerebral cortex from intrinsic
        functional connectivity mri. Cerebral cortex, 28(9):3095\u20133114,
        2018.
    """
    versions = ["fsaverage", "fsaverage5", "fsaverage6", "fslr32k"]
    if version not in versions:
        raise ValueError(
            f"The version of Schaefer et al., 2018 parcellation "
            f'requested "{version}" does not exist. Must be one of {versions}'
        )

    dataset_name = "atl-schaefer2018"
    _get_reference_info(dataset_name, verbose=verbose)

    keys = [f"{p}Parcels{n}Networks" for p in range(100, 1001, 100) for n in [7, 17]]

    fetched = fetch_file(
        dataset_name, keys=version, force=force, data_dir=data_dir, verbose=verbose
    )

    if version == "fslr32k":
        _fname = "atl-Schaefer2018_space-{}_hemi-{}_desc-{}_deterministic.dlabel.nii"
        data = {
            k: fetched / _fname.format(version, "LR", k)
            for k in keys
        }
    else:
        _fname = "atl-Schaefer2018_space-{}_hemi-{}_desc-{}_deterministic.annot"
        data = {
            k: SURFACE(
                fetched / _fname.format(version, "L", k),
                fetched / _fname.format(version, "R", k)
            )
            for k in keys
        }

    return Bunch(**data)


def fetch_mmpall(version="fslr32k", force=False, data_dir=None, verbose=1):
    """
    Download .label.gii files for Glasser et al., 2016 MMPAll atlas.

    This dataset contains

    If you used this data, please cite 1_.

    Parameters
    ----------
    version : {'fslr32k'}
        Specifies which surface annotation files should be matched to. Default:
        'fslr32k'

    Returns
    -------
    filenames : :class:`sklearn.utils.Bunch`
        Namedtuple with fields ('lh', 'rh') corresponding to filepaths to
        left/right hemisphere parcellation files

    Other Parameters
    ----------------
    force : bool, optional
        If True, will overwrite existing dataset. Default: False
    data_dir : str, optional
        Path to use as data directory. If not specified, will check for
        environmental variable 'NNT_DATA'; if that is not set, will use
        `~/nnt-data` instead. Default: None
    verbose : int, optional
        Modifies verbosity of download, where higher numbers mean more updates.
        Default: 1

    Notes
    -----
    License: https://www.humanconnectome.org/study/hcp-young-adult/document/wu-minn-hcp-consortium-open-access-data-use-terms

    References
    ----------
    .. [1] Matthew F Glasser, Timothy S Coalson, Emma C Robinson, Carl D Hacker,
        John Harwell, Essa Yacoub, Kamil Ugurbil, Jesper Andersson, Christian F
        Beckmann, Mark Jenkinson, and others. A multi-modal parcellation of
        human cerebral cortex. Nature, 536(7615):171\u2013178, 2016.
    """
    versions = ["fslr32k"]
    if version not in versions:
        raise ValueError(
            f"The version of Glasser et al., 2016 parcellation "
            f'requested "{version}" does not exist. Must be one of {versions}'
        )

    dataset_name = "atl-mmpall"
    _get_reference_info(dataset_name, verbose=verbose)

    fetched = fetch_file(
        dataset_name, keys=version, force=force, data_dir=data_dir, verbose=verbose
    )

    return SURFACE(
        fetched / f"atl-MMPAll_space-{version}_hemi-L_deterministic.label.gii",
        fetched / f"atl-MMPAll_space-{version}_hemi-R_deterministic.label.gii",
    )


def fetch_pauli2018(force=False, data_dir=None, verbose=1):
    """
    Download files for Pauli et al., 2018 subcortical parcellation.

    This dataset contains

    If you used this data, please cite 1_.

    Returns
    -------
    filenames : :class:`sklearn.utils.Bunch`
        Dictionary-like object with keys ['probabilistic', 'deterministic'],
        where corresponding values are filepaths to downloaded atlas files.

    Other Parameters
    ----------------
    force : bool, optional
        If True, will overwrite existing dataset. Default: False
    data_dir : str, optional
        Path to use as data directory. If not specified, will check for
        environmental variable 'NNT_DATA'; if that is not set, will use
        `~/nnt-data` instead. Default: None
    verbose : int, optional
        Modifies verbosity of download, where higher numbers mean more updates.
        Default: 1

    Notes
    -----
    License: CC-BY Attribution 4.0 International

    References
    ----------
    .. [1] Wolfgang M Pauli, Amanda N Nili, and J Michael Tyszka. A
        high-resolution probabilistic in vivo atlas of human subcortical brain
        nuclei. Scientific data, 5(1):1\u201313, 2018.
    """
    dataset_name = "atl-pauli2018"
    _get_reference_info(dataset_name, verbose=verbose)

    fetched = fetch_file(dataset_name, force=force, data_dir=data_dir, verbose=verbose)

    data = {
        "probabilistic": fetched
        / "atl-pauli2018_space-MNI152NLin2009cAsym_hemi-both_probabilistic.nii.gz",
        "deterministic": fetched
        / "atl-pauli2018_space-MNI152NLin2009cAsym_hemi-both_deterministic.nii.gz",
        "info": fetched / "atl-pauli2018_space-MNI152NLin2009cAsym_info.csv",
    }

    return Bunch(**data)


def fetch_tian2020msa():
    """Fetch Tian et al., 2020 subcortical parcellation."""
    pass


def fetch_voneconomo(force=False, data_dir=None, verbose=1):
    """
    Fetch von-Economo Koskinas probabilistic FreeSurfer atlas.

    This dataset contains

    If you used this data, please cite 1_.

    Returns
    -------
    filenames : :class:`sklearn.utils.Bunch`
        Dictionary-like object with keys ['gcs', 'ctab', 'info']

    Other Parameters
    ----------------
    force : bool, optional
        If True, will overwrite existing dataset. Default: False
    data_dir : str, optional
        Path to use as data directory. If not specified, will check for
        environmental variable 'NNT_DATA'; if that is not set, will use
        `~/nnt-data` instead. Default: None
    verbose : int, optional
        Modifies verbosity of download, where higher numbers mean more updates.
        Default: 1

    Notes
    -----
    License: CC-BY-NC-SA 4.0

    References
    ----------
    .. [1] Lianne H Scholtens, Marcel A de Reus, Siemon C de Lange, Ruben
        Schmidt, and Martijn P van den Heuvel. An mri von economo\u2013koskinas
        atlas. NeuroImage, 170:249\u2013256, 2018.
    """
    dataset_name = "atl-voneconomo_koskinas"
    _get_reference_info(dataset_name, verbose=verbose)

    fetched = fetch_file(dataset_name, force=force, data_dir=data_dir, verbose=verbose)

    data = {
        "gcs": SURFACE(
            fetched / "atl-vonEconomoKoskinas_hemi-L_probabilistic.gcs",
            fetched / "atl-vonEconomoKoskinas_hemi-R_probabilistic.gcs",
        ),
        "ctab": SURFACE(
            fetched / "atl-vonEconomoKoskinas_hemi-L_probabilistic.ctab",
            fetched / "atl-vonEconomoKoskinas_hemi-R_probabilistic.ctab",
        ),
        "info": fetched / "atl-vonEconomoKoskinas_info.csv",
    }

    return Bunch(**data)



"""Functions for generating randomized networks."""

import bct
import numpy as np
from tqdm import tqdm
from sklearn.utils.validation import check_random_state

from .. import has_numba
if has_numba:
    from numba import njit


def randmio_und(W, itr):
    """
    Optimized version of randmio_und.

    This function randomizes an undirected network, while preserving the
    degree distribution. The function does not preserve the strength
    distribution in weighted networks.

    This function is significantly faster if numba is enabled, because
    the main overhead is `np.random.randint`, see `here <https://stackoverflow.com/questions/58124646/why-in-python-is-random-randint-so-much-slower-than-random-random>`_

    Parameters
    ----------
    W : (N, N) array-like
        Undirected binary/weighted connection matrix
    itr : int
        rewiring parameter. Each edge is rewired approximately itr times.

    Returns
    -------
    W : (N, N) array-like
        Randomized network
    eff : int
        number of actual rewirings carried out
    """  # noqa: E501
    W = W.copy()
    n = len(W)
    i, j = np.where(np.triu(W > 0, 1))
    k = len(i)
    itr *= k

    # maximum number of rewiring attempts per iteration
    max_attempts = np.round(n * k / (n * (n - 1)))
    # actual number of successful rewirings
    eff = 0

    for _ in range(int(itr)):
        att = 0
        while att <= max_attempts:  # while not rewired
            while True:
                e1, e2 = np.random.randint(k), np.random.randint(k)
                while e1 == e2:
                    e2 = np.random.randint(k)
                a, b = i[e1], j[e1]
                c, d = i[e2], j[e2]

                if a != c and a != d and b != c and b != d:
                    break  # all 4 vertices must be different

            # flip edge c-d with 50% probability
            # to explore all potential rewirings
            if np.random.random() > 0.5:
                i[e2], j[e2] = d, c
                c, d = d, c

            # rewiring condition
            # not flipped
            # a--b    a  b
            #      TO  X
            # c--d    c  d
            # if flipped
            # a--b    a--b    a  b
            #      TO      TO  X
            # c--d    d--c    d  c
            if not (W[a, d] or W[c, b]):
                W[a, d] = W[a, b]
                W[a, b] = 0
                W[d, a] = W[b, a]
                W[b, a] = 0
                W[c, b] = W[c, d]
                W[c, d] = 0
                W[b, c] = W[d, c]
                W[d, c] = 0

                j[e1] = d
                j[e2] = b  # reassign edge indices
                eff += 1
                break
            att += 1

    return W, eff


if has_numba:
    randmio_und = njit(randmio_und)


def match_length_degree_distribution(
    W, D, nbins=10, nswap=1000, replacement=False, weighted=True, seed=None
):
    """
    Generate degree- and edge length-preserving surrogate connectomes.

    Parameters
    ----------
    W : (N, N) array-like
        weighted or binary symmetric connectivity matrix.
    D : (N, N) array-like
        symmetric distance matrix.
    nbins : int
        number of distance bins (edge length matrix is performed by swapping connections
        in the same bin). Default = 10.
    nswap : int
        total number of edge swaps to perform. Recommended = nnodes * 20.
        Default = 1000.
    replacement : bool, optional
        if True all the edges are available for swapping. Default = False.
    weighted : bool, optional
        if True the function returns a weighted matrix. Default = True.
    seed : float, optional
        Random seed. Default = None

    Returns
    -------
    newB : (N, N) array-like
        binary rewired matrix
    newW: (N, N) array-like
        weighted rewired matrix. Returns matrix of zeros if weighted=False.
    nr : int
        number of successful rewires

    Notes
    -----
    Takes a weighted, symmetric connectivity matrix `data` and Euclidean/fiber
    length matrix `distance` and generates a randomized network with:
        1. exactly the same degree sequence
        2. approximately the same edge length distribution
        3. exactly the same edge weight distribution
        4. approximately the same weight-length relationship

    Reference
    ---------
    Betzel, R. F., Bassett, D. S. (2018) Specificity and robustness of
    long-distance connections in weighted, interareal connectomes. PNAS.
    """
    rs = check_random_state(seed)
    N = len(W)
    # divide the distances by lengths
    bins = np.linspace(D[D.nonzero()].min(), D[D.nonzero()].max(), nbins + 1)
    bins[-1] += 1
    L = np.zeros((N, N))
    for n in range(nbins):
        i, j = np.where(np.logical_and(bins[n] <= D, D < bins[n + 1]))
        L[i, j] = n + 1

    # binarized connectivity
    B = (W != 0).astype(np.int_)

    # existing edges (only upper triangular cause it's symmetric)
    cn_x, cn_y = np.where(np.triu((B != 0) * B, k=1))

    tries = 0
    nr = 0
    newB = np.copy(B)

    while (len(cn_x) >= 2) & (nr < nswap):
        # choose randomly the edge to be rewired
        r = rs.randint(len(cn_x))
        n_x, n_y = cn_x[r], cn_y[r]
        tries += 1

        # options to rewire with
        # connected nodes that doesn't involve (n_x, n_y)
        index = (cn_x != n_x) & (cn_y != n_y) & (cn_y != n_x) & (cn_x != n_y)
        if len(np.where(index)[0]) == 0:
            cn_x = np.delete(cn_x, r)
            cn_y = np.delete(cn_y, r)

        else:
            ops1_x, ops1_y = cn_x[index], cn_y[index]
            # options that will preserve the distances
            # (ops1_x, ops1_y) such that
            # L(n_x,n_y) = L(n_x, ops1_x) & L(ops1_x,ops1_y) = L(n_y, ops1_y)
            index = (L[n_x, n_y] == L[n_x, ops1_x]) & (
                L[ops1_x, ops1_y] == L[n_y, ops1_y]
            )
            if len(np.where(index)[0]) == 0:
                cn_x = np.delete(cn_x, r)
                cn_y = np.delete(cn_y, r)

            else:
                ops2_x, ops2_y = ops1_x[index], ops1_y[index]
                # options of edges that didn't exist before
                index = [
                    (newB[min(n_x, ops2_x[i])][max(n_x, ops2_x[i])] == 0)
                    & (newB[min(n_y, ops2_y[i])][max(n_y, ops2_y[i])] == 0)
                    for i in range(len(ops2_x))
                ]
                if len(np.where(index)[0]) == 0:
                    cn_x = np.delete(cn_x, r)
                    cn_y = np.delete(cn_y, r)

                else:
                    ops3_x, ops3_y = ops2_x[index], ops2_y[index]

                    # choose randomly one edge from the final options
                    r1 = rs.randint(len(ops3_x))
                    nn_x, nn_y = ops3_x[r1], ops3_y[r1]

                    # Disconnect the existing edges
                    newB[n_x, n_y] = 0
                    newB[nn_x, nn_y] = 0
                    # Connect the new edges
                    newB[min(n_x, nn_x), max(n_x, nn_x)] = 1
                    newB[min(n_y, nn_y), max(n_y, nn_y)] = 1
                    # one successfull rewire!
                    nr += 1

                    # rewire with replacement
                    if replacement:
                        cn_x[r], cn_y[r] = min(n_x, nn_x), max(n_x, nn_x)
                        index = np.where((cn_x == nn_x) & (cn_y == nn_y))[0]
                        cn_x[index], cn_y[index] = min(n_y, nn_y), max(n_y, nn_y)
                    # rewire without replacement
                    else:
                        cn_x = np.delete(cn_x, r)
                        cn_y = np.delete(cn_y, r)
                        index = np.where((cn_x == nn_x) & (cn_y == nn_y))[0]
                        cn_x = np.delete(cn_x, index)
                        cn_y = np.delete(cn_y, index)

    if nr < nswap:
        print(f"I didn't finish, out of rewirable edges: {len(cn_x)}")

    i, j = np.triu_indices(N, k=1)
    # Make the connectivity matrix symmetric
    newB[j, i] = newB[i, j]

    # check the number of edges is preserved
    if len(np.where(B != 0)[0]) != len(np.where(newB != 0)[0]):
        print(
            f"ERROR --- number of edges changed, \
            B:{len(np.where(B != 0)[0])}, newB:{len(np.where(newB != 0)[0])}"
        )
    # check that the degree of the nodes it's the same
    for i in range(N):
        if np.sum(B[i]) != np.sum(newB[i]):
            print(
                f"ERROR --- node {i} changed k by: \
                {np.sum(B[i]) - np.sum(newB[i])}"
            )

    newW = np.zeros((N, N))
    if weighted:
        # Reassign the weights
        mask = np.triu(B != 0, k=1)
        inids = D[mask]
        iniws = W[mask]
        inids_index = np.argsort(inids)
        # Weights from the shortest to largest edges
        iniws = iniws[inids_index]
        mask = np.triu(newB != 0, k=1)
        finds = D[mask]
        i, j = np.where(mask)
        # Sort the new edges from the shortest to the largest
        finds_index = np.argsort(finds)
        i_sort = i[finds_index]
        j_sort = j[finds_index]
        # Assign the initial sorted weights
        newW[i_sort, j_sort] = iniws
        # Make it symmetrical
        newW[j_sort, i_sort] = iniws

    return newB, newW, nr


def strength_preserving_rand_sa(
    A,
    rewiring_iter=10,
    nstage=100,
    niter=10000,
    temp=1000,
    frac=0.5,
    energy_type="sse",
    energy_func=None,
    R=None,
    connected=None,
    verbose=False,
    seed=None,
):
    """
    Strength-preserving network randomization using simulated annealing.

    Randomize an undirected weighted network, while preserving
    the degree and strength sequences using simulated annealing.

    This function allows for a flexible choice of energy function.

    Parameters
    ----------
    A : (N, N) array-like
        Undirected weighted connectivity matrix
    rewiring_iter : int, optional
        Rewiring parameter. Default = 10.
        Each edge is rewired approximately rewiring_iter times.
    nstage : int, optional
        Number of annealing stages. Default = 100.
    niter : int, optional
        Number of iterations per stage. Default = 10000.
    temp : float, optional
        Initial temperature. Default = 1000.
    frac : float, optional
        Fractional decrease in temperature per stage. Default = 0.5.
    energy_type: str, optional
        Energy function to minimize. Can be either:
            'sse': Sum of squared errors between strength sequence vectors
                   of the original network and the randomized network
            'max': Maximum absolute error
            'mae': Mean absolute error
            'mse': Mean squared error
            'rmse': Root mean squared error
        Default = 'sse'.
    energy_func: callable, optional
        Callable with two positional arguments corresponding to
        two strength sequence numpy arrays that returns an energy value.
        Overwrites “energy_type”.
        See “energy_type” for specifying a predefined energy type instead.
    R : (N, N) array-like, optional
        Pre-randomized connectivity matrix.
        If None, a rewired connectivity matrix is generated using the
        Maslov & Sneppen algorithm.
        Default = None.
    connected: bool, optional
        Whether to ensure connectedness of the randomized network.
        By default, this is inferred from data.
    verbose: bool, optional
        Whether to print status to screen at the end of every stage.
        Default = False.
    seed: float, optional
        Random seed. Default = None.

    Returns
    -------
    B : (N, N) array-like
        Randomized connectivity matrix
    min_energy : float
        Minimum energy obtained by annealing

    Notes
    -----
    Uses Maslov & Sneppen rewiring model to produce a
    surrogate connectivity matrix, B, with the same
    size, density, and degree sequence as A.
    The weights are then permuted to optimize the
    match between the strength sequences of A and B
    using simulated annealing.

    This function is adapted from a function written in MATLAB
    by Richard Betzel.

    References
    ----------
    Misic, B. et al. (2015) Cooperative and Competitive Spreading Dynamics
    on the Human Connectome. Neuron.
    Milisav, F. et al. (2024) A simulated annealing algorithm for
    randomizing weighted networks.
    """
    try:
        A = np.asarray(A)
    except TypeError as err:
        msg = "A must be array_like. Received: {}.".format(type(A))
        raise TypeError(msg) from err

    if frac > 1 or frac <= 0:
        msg = "frac must be between 0 and 1. " "Received: {}.".format(frac)
        raise ValueError(msg)

    rs = check_random_state(seed)

    n = A.shape[0]
    s = np.sum(A, axis=1)  # strengths of A

    # Maslov & Sneppen rewiring
    if R is None:
        # ensuring connectedness if the original network is connected
        if connected is None:
            connected = False if bct.number_of_components(A) > 1 else True
        if connected:
            B = bct.randmio_und_connected(A, rewiring_iter, seed=seed)[0]
        else:
            B = bct.randmio_und(A, rewiring_iter, seed=seed)[0]
    else:
        B = R.copy()

    u, v = np.triu(B, k=1).nonzero()  # upper triangle indices
    wts = np.triu(B, k=1)[(u, v)]  # upper triangle values
    m = len(wts)
    sb = np.sum(B, axis=1)  # strengths of B

    if energy_func is not None:
        energy = energy_func(s, sb)
    elif energy_type == "sse":
        energy = np.sum((s - sb) ** 2)
    elif energy_type == "max":
        energy = np.max(np.abs(s - sb))
    elif energy_type == "mae":
        energy = np.mean(np.abs(s - sb))
    elif energy_type == "mse":
        energy = np.mean((s - sb) ** 2)
    elif energy_type == "rmse":
        energy = np.sqrt(np.mean((s - sb) ** 2))
    else:
        msg = (
            "energy_type must be one of 'sse', 'max', "
            "'mae', 'mse', or 'rmse'. Received: {}.".format(energy_type)
        )
        raise ValueError(msg)

    energymin = energy
    wtsmin = wts.copy()

    if verbose:
        print("\ninitial energy {:.5f}".format(energy))

    for istage in tqdm(range(nstage), desc="annealing progress"):
        naccept = 0
        for _ in range(niter):
            # permutation
            e1 = rs.randint(m)
            e2 = rs.randint(m)

            a, b = u[e1], v[e1]
            c, d = u[e2], v[e2]

            sb_prime = sb.copy()
            sb_prime[[a, b]] = sb_prime[[a, b]] - wts[e1] + wts[e2]
            sb_prime[[c, d]] = sb_prime[[c, d]] + wts[e1] - wts[e2]

            if energy_func is not None:
                energy_prime = energy_func(sb_prime, s)
            elif energy_type == "sse":
                energy_prime = np.sum((sb_prime - s) ** 2)
            elif energy_type == "max":
                energy_prime = np.max(np.abs(sb_prime - s))
            elif energy_type == "mae":
                energy_prime = np.mean(np.abs(sb_prime - s))
            elif energy_type == "mse":
                energy_prime = np.mean((sb_prime - s) ** 2)
            elif energy_type == "rmse":
                energy_prime = np.sqrt(np.mean((sb_prime - s) ** 2))
            else:
                msg = (
                    "energy_type must be one of 'sse', 'max', "
                    "'mae', 'mse', or 'rmse'. "
                    "Received: {}.".format(energy_type)
                )
                raise ValueError(msg)

            # permutation acceptance criterion
            if energy_prime < energy or rs.rand() < np.exp(
                -(energy_prime - energy) / temp
            ):
                sb = sb_prime.copy()
                wts[[e1, e2]] = wts[[e2, e1]]
                energy = energy_prime
                if energy < energymin:
                    energymin = energy
                    wtsmin = wts.copy()
                naccept = naccept + 1

        # temperature update
        temp = temp * frac
        if verbose:
            print(
                "\nstage {:d}, temp {:.5f}, best energy {:.5f}, "
                "frac of accepted moves {:.3f}".format(
                    istage, temp, energymin, naccept / niter
                )
            )

    B = np.zeros((n, n))
    B[(u, v)] = wtsmin
    B = B + B.T

    return B, energymin


def strength_preserving_rand_sa_mse_opt(
    A,
    rewiring_iter=10,
    nstage=100,
    niter=10000,
    temp=1000,
    frac=0.5,
    R=None,
    connected=None,
    verbose=False,
    seed=None,
):
    """
    Strength-preserving network randomization using simulated annealing.

    Randomize an undirected weighted network, while preserving
    the degree and strength sequences using simulated annealing.

    This function has been optimized for speed but only allows the
    mean squared error energy function.

    Parameters
    ----------
    A : (N, N) array-like
        Undirected weighted connectivity matrix
    rewiring_iter : int, optional
        Rewiring parameter. Default = 10.
        Each edge is rewired approximately rewiring_iter times.
    nstage : int, optional
        Number of annealing stages. Default = 100.
    niter : int, optional
        Number of iterations per stage. Default = 10000.
    temp : float, optional
        Initial temperature. Default = 1000.
    frac : float, optional
        Fractional decrease in temperature per stage. Default = 0.5.
    R : (N, N) array-like, optional
        Pre-randomized connectivity matrix.
        If None, a rewired connectivity matrix is generated using the
        Maslov & Sneppen algorithm.
        Default = None.
    connected: bool, optional
        Whether to ensure connectedness of the randomized network.
        By default, this is inferred from data.
    verbose: bool, optional
        Whether to print status to screen at the end of every stage.
        Default = False.
    seed: float, optional
        Random seed. Default = None.

    Returns
    -------
    B : (N, N) array-like
        Randomized connectivity matrix
    min_energy : float
        Minimum energy obtained by annealing

    Notes
    -----
    Uses Maslov & Sneppen rewiring model to produce a
    surrogate connectivity matrix, B, with the same
    size, density, and degree sequence as A.
    The weights are then permuted to optimize the
    match between the strength sequences of A and B
    using simulated annealing.

    This function is adapted from a function written in MATLAB
    by Richard Betzel and was optimized by Vincent Bazinet.

    References
    ----------
    Misic, B. et al. (2015) Cooperative and Competitive Spreading Dynamics
    on the Human Connectome. Neuron.
    Milisav, F. et al. (2024) A simulated annealing algorithm for
    randomizing weighted networks.
    """
    try:
        A = np.asarray(A)
    except TypeError as err:
        msg = "A must be array_like. Received: {}.".format(type(A))
        raise TypeError(msg) from err

    if frac > 1 or frac <= 0:
        msg = "frac must be between 0 and 1. " "Received: {}.".format(frac)
        raise ValueError(msg)

    rs = check_random_state(seed)

    n = A.shape[0]
    s = np.sum(A, axis=1)  # strengths of A

    # Maslov & Sneppen rewiring
    if R is None:
        # ensuring connectedness if the original network is connected
        if connected is None:
            connected = False if bct.number_of_components(A) > 1 else True
        if connected:
            B = bct.randmio_und_connected(A, rewiring_iter, seed=seed)[0]
        else:
            B = bct.randmio_und(A, rewiring_iter, seed=seed)[0]
    else:
        B = R.copy()

    u, v = np.triu(B, k=1).nonzero()  # upper triangle indices
    wts = np.triu(B, k=1)[(u, v)]  # upper triangle values
    m = len(wts)
    sb = np.sum(B, axis=1)  # strengths of B

    energy = np.mean((s - sb) ** 2)

    energymin = energy
    wtsmin = wts.copy()

    if verbose:
        print("\ninitial energy {:.5f}".format(energy))

    for istage in tqdm(range(nstage), desc="annealing progress"):
        naccept = 0
        for (e1, e2), prob in zip(rs.randint(m, size=(niter, 2)), rs.rand(niter)):
            # permutation
            a, b, c, d = u[e1], v[e1], u[e2], v[e2]
            wts_change = wts[e1] - wts[e2]
            delta_energy = (
                2
                * wts_change
                * (
                    2 * wts_change
                    + (s[a] - sb[a])
                    + (s[b] - sb[b])
                    - (s[c] - sb[c])
                    - (s[d] - sb[d])
                )
            ) / n

            # permutation acceptance criterion
            if delta_energy < 0 or prob < np.e ** (-(delta_energy) / temp):
                sb[[a, b]] -= wts_change
                sb[[c, d]] += wts_change
                wts[[e1, e2]] = wts[[e2, e1]]

                energy = np.mean((sb - s) ** 2)

                if energy < energymin:
                    energymin = energy
                    wtsmin = wts.copy()
                naccept = naccept + 1

        # temperature update
        temp = temp * frac
        if verbose:
            print(
                "\nstage {:d}, temp {:.5f}, best energy {:.5f}, "
                "frac of accepted moves {:.3f}".format(
                    istage, temp, energymin, naccept / niter
                )
            )

    B = np.zeros((n, n))
    B[(u, v)] = wtsmin
    B = B + B.T

    return B, energymin


def strength_preserving_rand_sa_dir(
    A,
    rewiring_iter=10,
    nstage=100,
    niter=10000,
    temp=1000,
    frac=0.5,
    energy_type="sse",
    energy_func=None,
    connected=True,
    verbose=False,
    seed=None,
):
    """
    Strength-preserving network randomization using simulated annealing.

    Randomize a directed weighted network, while preserving
    the in- and out-degree and strength sequences using simulated annealing.

    Parameters
    ----------
    A : (N, N) array-like
        Directed weighted connectivity matrix
    rewiring_iter : int, optional
        Rewiring parameter. Default = 10.
        Each edge is rewired approximately rewiring_iter times.
    nstage : int, optional
        Number of annealing stages. Default = 100.
    niter : int, optional
        Number of iterations per stage. Default = 10000.
    temp : float, optional
        Initial temperature. Default = 1000.
    frac : float, optional
        Fractional decrease in temperature per stage. Default = 0.5.
    energy_type: str, optional
        Energy function to minimize. Can be either:
            'sse': Sum of squared errors between strength sequence vectors
                   of the original network and the randomized network
            'max': Maximum absolute error
            'mae': Mean absolute error
            'mse': Mean squared error
            'rmse': Root mean squared error
        Default = 'sse'.
    energy_func: callable, optional
        Callable with two positional arguments corresponding to
        two strength sequence numpy arrays that returns an energy value.
        Overwrites “energy_type”.
        See “energy_type” for specifying a predefined energy type instead.
    connected: bool, optional
        Whether to ensure connectedness of the randomized network.
        Default = True.
    verbose: bool, optional
        Whether to print status to screen at the end of every stage.
        Default = False.
    seed: float, optional
        Random seed. Default = None.

    Returns
    -------
    B : (N, N) array-like
        Randomized connectivity matrix
    min_energy : float
        Minimum energy obtained by annealing

    Notes
    -----
    Uses Maslov & Sneppen rewiring model to produce a
    surrogate connectivity matrix, B, with the same
    size, density, and in- and out-degree sequences as A.
    The weights are then permuted to optimize the
    match between the strength sequences of A and B
    using simulated annealing.
    Both in- and out-strengths are preserved.

    This function is adapted from a function written in MATLAB
    by Richard Betzel.

    References
    ----------
    Misic, B. et al. (2015) Cooperative and Competitive Spreading Dynamics
    on the Human Connectome. Neuron.
    Rubinov, M. (2016) Constraints and spandrels of interareal connectomes.
    Nature Communications.
    Milisav, F. et al. (2024) A simulated annealing algorithm for
    randomizing weighted networks.
    """
    try:
        A = np.asarray(A)
    except TypeError as err:
        msg = "A must be array_like. Received: {}.".format(type(A))
        raise TypeError(msg) from err

    if frac > 1 or frac <= 0:
        msg = "frac must be between 0 and 1. " "Received: {}.".format(frac)
        raise ValueError(msg)

    rs = check_random_state(seed)

    n = A.shape[0]
    s_in = np.sum(A, axis=0)  # in-strengths of A
    s_out = np.sum(A, axis=1)  # out-strengths of A

    # Maslov & Sneppen rewiring
    if connected:
        B = bct.randmio_dir_connected(A, rewiring_iter, seed=seed)[0]
    else:
        B = bct.randmio_dir(A, rewiring_iter, seed=seed)[0]

    u, v = B.nonzero()  # nonzero indices of B
    wts = B[(u, v)]  # nonzero values of B
    m = len(wts)
    sb_in = np.sum(B, axis=0)  # in-strengths of B
    sb_out = np.sum(B, axis=1)  # out-strengths of B

    if energy_func is not None:
        energy = energy_func(s_in, sb_in) + energy_func(s_out, sb_out)
    elif energy_type == "sse":
        energy = np.sum((s_in - sb_in) ** 2) + np.sum((s_out - sb_out) ** 2)
    elif energy_type == "max":
        energy = np.max(np.abs(s_in - sb_in)) + np.max(np.abs(s_out - sb_out))
    elif energy_type == "mae":
        energy = np.mean(np.abs(s_in - sb_in)) + np.mean(np.abs(s_out - sb_out))
    elif energy_type == "mse":
        energy = np.mean((s_in - sb_in) ** 2) + np.mean((s_out - sb_out) ** 2)
    elif energy_type == "rmse":
        energy = np.sqrt(np.mean((s_in - sb_in) ** 2)) + np.sqrt(
            np.mean((s_out - sb_out) ** 2)
        )
    else:
        msg = (
            "energy_type must be one of 'sse', 'max', "
            "'mae', 'mse', or 'rmse'. Received: {}.".format(energy_type)
        )
        raise ValueError(msg)

    energymin = energy
    wtsmin = wts.copy()

    if verbose:
        print("\ninitial energy {:.5f}".format(energy))

    for istage in tqdm(range(nstage), desc="annealing progress"):
        naccept = 0
        for _ in range(niter):
            # permutation
            e1 = rs.randint(m)
            e2 = rs.randint(m)

            a, b = u[e1], v[e1]
            c, d = u[e2], v[e2]

            sb_prime_in = sb_in.copy()
            sb_prime_out = sb_out.copy()
            sb_prime_in[b] = sb_prime_in[b] - wts[e1] + wts[e2]
            sb_prime_out[a] = sb_prime_out[a] - wts[e1] + wts[e2]
            sb_prime_in[d] = sb_prime_in[d] - wts[e2] + wts[e1]
            sb_prime_out[c] = sb_prime_out[c] - wts[e2] + wts[e1]

            if energy_func is not None:
                energy_prime = energy_func(sb_prime_in, s_in) + energy_func(
                    sb_prime_out, s_out
                )
            elif energy_type == "sse":
                energy_prime = np.sum((sb_prime_in - s_in) ** 2) + np.sum(
                    (sb_prime_out - s_out) ** 2
                )
            elif energy_type == "max":
                energy_prime = np.max(np.abs(sb_prime_in - s_in)) + np.max(
                    np.abs(sb_prime_out - s_out)
                )
            elif energy_type == "mae":
                energy_prime = np.mean(np.abs(sb_prime_in - s_in)) + np.mean(
                    np.abs(sb_prime_out - s_out)
                )
            elif energy_type == "mse":
                energy_prime = np.mean((sb_prime_in - s_in) ** 2) + np.mean(
                    (sb_prime_out - s_out) ** 2
                )
            elif energy_type == "rmse":
                energy_prime = np.sqrt(np.mean((sb_prime_in - s_in) ** 2)) + np.sqrt(
                    np.mean((sb_prime_out - s_out) ** 2)
                )
            else:
                msg = (
                    "energy_type must be one of 'sse', 'max', "
                    "'mae', 'mse', or 'rmse'. "
                    "Received: {}.".format(energy_type)
                )
                raise ValueError(msg)

            # permutation acceptance criterion
            if energy_prime < energy or rs.rand() < np.exp(
                -(energy_prime - energy) / temp
            ):
                sb_in = sb_prime_in.copy()
                sb_out = sb_prime_out.copy()
                wts[[e1, e2]] = wts[[e2, e1]]
                energy = energy_prime
                if energy < energymin:
                    energymin = energy
                    wtsmin = wts.copy()
                naccept = naccept + 1

        # temperature update
        temp = temp * frac
        if verbose:
            print(
                "\nstage {:d}, temp {:.5f}, best energy {:.5f}, "
                "frac of accepted moves {:.3f}".format(
                    istage, temp, energymin, naccept / niter
                )
            )

    B = np.zeros((n, n))
    B[(u, v)] = wtsmin

    return B, energymin



"""For testing netneurotools.spatial.spatial_stats functionality."""

import pytest
import numpy as np

from netneurotools.spatial.spatial_stats import (
    morans_i, local_morans_i,
    gearys_c, local_gearys_c,
    lees_l, local_lees_l
)


@pytest.fixture(scope="session")
def generate_test_data(request):
    """Generate test data for spatial statistics tests."""
    n = request.param["n"]
    which = request.param["which"]

    rng = np.random.default_rng(1234)
    annot_1 = rng.random(n)
    annot_2 = rng.random(n)
    weight = rng.random((n, n))

    expected_all = {
        "morans_i": {3: 0.04811638, 4: 0.10194804},
        "local_morans_i": {
            3: [0.17645791, 0.03789162, -0.04622557],
            4: [0.40695548, -0.04131271, -0.17433353, 0.52496365],
        },
        "gearys_c": {3: 0.5665884, 4: 0.6756672},
        "local_gearys_c": {
            3: [1.06577412558015, 2.51230240958039, 2.36109222974053],
            4: [
                3.0202751462928, 3.82737772381897,
                3.33112622823278, 2.48027429966568
            ]
        },
        "lees_l": {3: -0.03207622, 4: 0.023249778},
        "local_lees_l": {
            3: [-0.297374681957742, 0.0116826474633146, -0.0293788020084983],
            4: [
                0.439955191359293, -0.0105931628916757,
                -0.253241702688354, 0.351131157157366
            ]
        }
    }

    return annot_1, annot_2, weight, expected_all[which][n]


@pytest.mark.parametrize(
    "generate_test_data",
    [
        pytest.param({"n": n, "which": "morans_i"}, id=f"morans_i-{n}")
        for n in [3, 4]
    ],
    indirect=True,
)
def test_morans_i(generate_test_data):
    """Test Moran's I calculation."""
    annot_1, _, weight, expected = generate_test_data
    assert np.isclose(morans_i(annot_1, weight, use_numba=False), expected)
    assert np.isclose(morans_i(annot_1, weight, use_numba=True), expected)


@pytest.mark.parametrize(
    "generate_test_data",
    [
        pytest.param({"n": n, "which": "local_morans_i"}, id=f"local_morans_i-{n}")
        for n in [3, 4]
    ],
    indirect=True,
)
def test_local_morans_i(generate_test_data):
    """Test local Moran's I calculation."""
    annot_1, _, weight, expected = generate_test_data
    assert np.allclose(local_morans_i(annot_1, weight), expected)


@pytest.mark.parametrize(
    "generate_test_data",
    [
        pytest.param({"n": n, "which": "gearys_c"}, id=f"gearys_c-{n}")
        for n in [3, 4]
    ],
    indirect=True,
)
def test_gearys_c(generate_test_data):
    """Test Geary's C calculation."""
    annot_1, _, weight, expected = generate_test_data
    assert np.isclose(gearys_c(annot_1, weight, use_numba=False), expected)
    assert np.isclose(gearys_c(annot_1, weight, use_numba=True), expected)


@pytest.mark.parametrize(
    "generate_test_data",
    [
        pytest.param({"n": n, "which": "local_gearys_c"}, id=f"local_gearys_c-{n}")
        for n in [3, 4]
    ],
    indirect=True,
)
def test_local_gearys_c(generate_test_data):
    """Test local Geary's C calculation."""
    annot_1, _, weight, expected = generate_test_data
    assert np.allclose(local_gearys_c(annot_1, weight), expected)


@pytest.mark.parametrize(
    "generate_test_data",
    [
        pytest.param({"n": n, "which": "lees_l"}, id=f"lees_l-{n}")
        for n in [3, 4]
    ],
    indirect=True,
)
def test_lees_l(generate_test_data):
    """Test Lee's L calculation."""
    annot_1, annot_2, weight, expected = generate_test_data
    assert np.isclose(lees_l(annot_1, annot_2, weight, use_numba=False), expected)
    assert np.isclose(lees_l(annot_1, annot_2, weight, use_numba=True), expected)


@pytest.mark.parametrize(
    "generate_test_data",
    [
        pytest.param({"n": n, "which": "local_lees_l"}, id=f"local_lees_l-{n}")
        for n in [3, 4]
    ],
    indirect=True,
)
def test_local_lees_l(generate_test_data):
    """Test local Lee's L calculation."""
    annot_1, annot_2, weight, expected = generate_test_data
    assert np.allclose(local_lees_l(annot_1, annot_2, weight), expected)



"""For testing netneurotools.datasets.fetch_* functionality."""

import os
import pytest
from pathlib import Path
import numpy as np
from netneurotools import datasets


@pytest.mark.fetcher
class TestFetchTemplate:
    """Test fetching of template datasets."""

    @pytest.mark.parametrize(
        "version", ["fsaverage", "fsaverage3", "fsaverage4", "fsaverage5", "fsaverage6"]
    )
    def test_fetch_fsaverage(self, tmp_path, version):
        """Test fetching of fsaverage surfaces."""
        fsaverage = datasets.fetch_fsaverage(
            version=version, data_dir=tmp_path, verbose=0
        )
        for k in ["orig", "white", "smoothwm", "pial", "inflated", "sphere"]:
            assert k in fsaverage
            assert fsaverage[k].L.exists()
            assert fsaverage[k].R.exists()

    @pytest.mark.parametrize(
        "version", ["fsaverage", "fsaverage4", "fsaverage5", "fsaverage6"]
    )
    def test_fetch_fsaverage_curated(self, tmp_path, version):
        """Test fetching of curated fsaverage surfaces."""
        fsaverage = datasets.fetch_fsaverage_curated(
            version=version, data_dir=tmp_path, verbose=0
        )
        for k in ["white", "pial", "inflated", "sphere", "medial", "sulc", "vaavg"]:
            assert k in fsaverage
            assert fsaverage[k].L.exists()
            assert fsaverage[k].R.exists()

    def test_fetch_hcp_standards(self, tmp_path):
        """Test fetching of HCP standard meshes."""
        hcp = datasets.fetch_hcp_standards(data_dir=tmp_path, verbose=0)
        assert hcp.exists()

    @pytest.mark.parametrize("version", ["fslr4k", "fslr8k", "fslr32k", "fslr164k"])
    def test_fetch_fslr_curated(self, tmp_path, version):
        """Test fetching of curated fsLR surfaces."""
        fslr = datasets.fetch_fslr_curated(
            version=version, data_dir=tmp_path, verbose=0
        )
        for k in [
            "midthickness",
            "inflated",
            "veryinflated",
            "sphere",
            "medial",
            "sulc",
            "vaavg",
        ]:
            if version in ["fslr4k", "fslr8k"] and k == "veryinflated":
                continue
            assert k in fslr
            assert fslr[k].L.exists()
            assert fslr[k].R.exists()

    @pytest.mark.parametrize("version", ["v1", "v2"])
    def test_fetch_civet(self, tmp_path, version):
        """Test fetching of CIVET templates."""
        civet = datasets.fetch_civet(version=version, data_dir=tmp_path, verbose=0)
        for key in ("mid", "white"):
            assert key in civet
            assert civet[key].L.exists()
            assert civet[key].R.exists()

    @pytest.mark.parametrize("version", ["civet41k", "civet164k"])
    def test_fetch_civet_curated(self, tmp_path, version):
        """Test fetching of curated CIVET templates."""
        civet = datasets.fetch_civet_curated(
            version=version, data_dir=tmp_path, verbose=0
        )
        for k in [
            "white",
            "midthickness",
            "inflated",
            "veryinflated",
            "sphere",
            "medial",
            "sulc",
            "vaavg",
        ]:
            assert k in civet
            assert civet[k].L.exists()
            assert civet[k].R.exists()

    def test_fetch_conte69(self, tmp_path):
        """Test fetching of Conte69 surfaces."""
        conte = datasets.fetch_conte69(data_dir=tmp_path, verbose=0)
        assert all(
            hasattr(conte, k) for k in ["midthickness", "inflated", "vinflated", "info"]
        )

    def test_fetch_yerkes19(self, tmp_path):
        """Test fetching of Yerkes19 surfaces."""
        yerkes19 = datasets.fetch_yerkes19(data_dir=tmp_path, verbose=0)
        assert all(
            hasattr(yerkes19, k) for k in ["midthickness", "inflated", "vinflated"]
        )


@pytest.mark.fetcher
class TestFetchAtlas:
    """Test fetching of atlas datasets."""

    @pytest.mark.parametrize(
        "version, expected",
        [
            ("MNI152NLin2009aSym", [1, 1, 1, 1, 1]),
            ("fsaverage", [2, 2, 2, 2, 2]),
            ("fsaverage5", [2, 2, 2, 2, 2]),
            ("fsaverage6", [2, 2, 2, 2, 2]),
            ("fslr32k", [2, 2, 2, 2, 2]),
            ("gcs", [2, 2, 2, 2, 6]),
        ],
    )
    def test_fetch_cammoun2012(self, tmp_path, version, expected):
        """Test fetching of Cammoun2012 parcellations."""
        keys = ["scale033", "scale060", "scale125", "scale250", "scale500"]
        cammoun = datasets.fetch_cammoun2012(version, data_dir=tmp_path, verbose=0)

        # output has expected keys
        assert all(hasattr(cammoun, k) for k in keys)
        # and keys are expected lengths!
        for k, e in zip(keys, expected):
            out = getattr(cammoun, k)
            if isinstance(out, (tuple, list)):
                assert len(out) == e
            else:
                assert isinstance(out, Path) and str(out).endswith(".nii.gz")

    @pytest.mark.parametrize(
        "version", ["fsaverage", "fsaverage5", "fsaverage6", "fslr32k"]
    )
    def test_fetch_schaefer2018(self, tmp_path, version):
        """Test fetching of Schaefer2018 parcellations."""
        keys = [
            f"{p}Parcels{n}Networks" for p in range(100, 1001, 100) for n in [7, 17]
        ]
        schaefer = datasets.fetch_schaefer2018(version, data_dir=tmp_path, verbose=0)

        if version == "fslr32k":
            assert all(k in schaefer and os.path.isfile(schaefer[k]) for k in keys)
        else:
            for k in keys:
                assert k in schaefer
                assert len(schaefer[k]) == 2
                assert all(os.path.isfile(hemi) for hemi in schaefer[k])

    def test_fetch_mmpall(self, tmp_path):
        """Test fetching of MMPAll parcellations."""
        mmp = datasets.fetch_mmpall(data_dir=tmp_path, verbose=0)
        assert len(mmp) == 2
        assert all(os.path.isfile(hemi) for hemi in mmp)
        assert all(hasattr(mmp, attr) for attr in ("L", "R"))

    def test_fetch_pauli2018(self, tmp_path):
        """Test fetching of Pauli2018 parcellations."""
        pauli = datasets.fetch_pauli2018(data_dir=tmp_path, verbose=0)
        assert all(
            hasattr(pauli, k) and os.path.isfile(pauli[k])
            for k in ["probabilistic", "deterministic", "info"]
        )

    @pytest.mark.xfail
    def test_fetch_tian2020msa(self, tmp_path):
        """Test fetching of tian2020msa parcellations."""
        assert False

    def test_fetch_voneconomo(self, tmp_path):
        """Test fetching of von Economo parcellations."""
        vek = datasets.fetch_voneconomo(data_dir=tmp_path, verbose=0)
        assert all(hasattr(vek, k) and len(vek[k]) == 2 for k in ["gcs", "ctab"])
        assert isinstance(vek.get("info"), Path)


@pytest.mark.fetcher
class TestFetchProject:
    """Test fetching of project datasets."""

    def test_fetch_vazquez_rodriguez2019(self, tmp_path):
        """Test fetching of Vazquez-Rodriguez2019 dataset."""
        vazquez = datasets.fetch_vazquez_rodriguez2019(data_dir=tmp_path, verbose=0)
        for k in ["rsquared", "gradient"]:
            assert hasattr(vazquez, k)
            assert isinstance(getattr(vazquez, k), np.ndarray)

    @pytest.mark.xfail
    def test_fetch_mirchi2018(self, tmp_path):
        """Test fetching of Mirchi2018 dataset."""
        X, Y = datasets.fetch_mirchi2018(data_dir=tmp_path, verbose=0)
        assert isinstance(X, np.ndarray)
        assert X.shape == (73, 198135)
        assert isinstance(Y, np.ndarray)
        assert Y.shape == (73, 13)

    def test_fetch_hansen_manynetworks(self, tmp_path):
        """Test fetching of Hansen et al., 2023 many-networks dataset."""
        hansen = datasets.fetch_hansen_manynetworks(data_dir=tmp_path, verbose=0)
        assert hansen.exists()
        # assert "cammoun033" in hansen
        # assert "gene" in hansen["cammoun033"]
        # assert isinstance(hansen["cammoun033"]["gene"], Path)

    def test_fetch_hansen_receptors(self, tmp_path):
        """Test fetching of Hansen et al., 2022 receptor dataset."""
        hansen = datasets.fetch_hansen_receptors(data_dir=tmp_path, verbose=0)
        assert hansen.exists()

    def test_fetch_hansen_genescognition(self, tmp_path):
        """Test fetching of Hansen et al., 2021 gene-cognition dataset."""
        hansen = datasets.fetch_hansen_genescognition(data_dir=tmp_path, verbose=0)
        assert hansen.exists()

    def test_fetch_hansen_brainstemfc(self, tmp_path):
        """Test fetching of Hansen et al., 2024 brainstem dataset."""
        hansen = datasets.fetch_hansen_brainstemfc(data_dir=tmp_path, verbose=0)
        assert hansen.exists()

    def test_fetch_shafiei_megfmrimapping(self, tmp_path):
        """Test fetching of Shafiei et al., 2022 & 2023 HCP-MEG dataset."""
        shafiei = datasets.fetch_shafiei_megfmrimapping(data_dir=tmp_path, verbose=0)
        assert shafiei.exists()

    def test_fetch_shafiei_megdynamics(self, tmp_path):
        """Test fetching of Shafiei et al., 2022 & 2023 HCP-MEG dataset."""
        shafiei = datasets.fetch_shafiei_megdynamics(data_dir=tmp_path, verbose=0)
        assert shafiei.exists()

    def test_fetch_suarez_mami(self, tmp_path):
        """Test fetching of Suarez et al., 2022 mami dataset."""
        suarez = datasets.fetch_suarez_mami(data_dir=tmp_path, verbose=0)
        assert suarez.exists()

    @pytest.mark.parametrize(
        "dataset, expected",
        [
            ("celegans", ["conn", "dist", "labels", "ref"]),
            ("drosophila", ["conn", "coords", "labels", "networks", "ref"]),
            ("human_func_scale033", ["conn", "coords", "labels", "ref"]),
            ("human_func_scale060", ["conn", "coords", "labels", "ref"]),
            ("human_func_scale125", ["conn", "coords", "labels", "ref"]),
            ("human_func_scale250", ["conn", "coords", "labels", "ref"]),
            ("human_func_scale500", ["conn", "coords", "labels", "ref"]),
            ("human_struct_scale033", ["conn", "coords", "dist", "labels", "ref"]),
            ("human_struct_scale060", ["conn", "coords", "dist", "labels", "ref"]),
            ("human_struct_scale125", ["conn", "coords", "dist", "labels", "ref"]),
            ("human_struct_scale250", ["conn", "coords", "dist", "labels", "ref"]),
            ("human_struct_scale500", ["conn", "coords", "dist", "labels", "ref"]),
            ("macaque_markov", ["conn", "dist", "labels", "ref"]),
            ("macaque_modha", ["conn", "coords", "dist", "labels", "ref"]),
            ("mouse", ["acronyms", "conn", "coords", "dist", "labels", "ref"]),
            ("rat", ["conn", "labels", "ref"]),
        ],
    )
    def test_fetch_famous_gmat(self, tmp_path, dataset, expected):
        """Test fetching of famous G.mat datasets."""
        connectome = datasets.fetch_famous_gmat(dataset, data_dir=tmp_path, verbose=0)

        expected.remove("ref")
        for key in expected:
            assert key in connectome
            assert isinstance(connectome[key], str if key == "ref" else np.ndarray)



"""For testing netneurotools.plotting.mpl_plotters functionality."""

import numpy as np
import matplotlib.pyplot as plt
from netneurotools import plotting


def test_grid_communities():
    """Test _grid_communities function."""
    comms = np.asarray([0, 0, 0, 0, 1, 1, 1, 1, 2, 2])
    # check that comms with / without 0 community label yields same output
    assert np.allclose(plotting._grid_communities(comms), [0, 4, 8, 10])
    assert np.allclose(plotting._grid_communities(comms + 1), [0, 4, 8, 10])


def test_sort_communities():
    """Test sort_communities function."""
    data = np.arange(9).reshape(3, 3)
    comms = np.asarray([0, 0, 2])
    # check that comms with / without 0 community label yields same output
    assert np.allclose(plotting._sort_communities(data, comms), [1, 0, 2])
    assert np.allclose(plotting._sort_communities(data, comms + 1), [1, 0, 2])


def test_plot_mod_heatmap():
    """Test plot_mod_heatmap function."""
    data = np.random.rand(100, 100)
    comms = np.random.choice(4, size=(100,))
    ax = plotting.plot_mod_heatmap(data, comms)
    assert isinstance(ax, plt.Axes)


def test_plot_point_brain():
    """Test plot_point_brain function."""
    data = np.random.rand(100)
    coords = np.random.rand(100, 3)
    fig, axes = plotting.plot_point_brain(data, coords)
    assert isinstance(fig, plt.Figure)
    assert isinstance(axes, np.ndarray)



"""Functions for handling spatial brain data."""


from .spatial_stats import (
    morans_i, local_morans_i,
    gearys_c, local_gearys_c,
    lees_l, local_lees_l
)


__all__ = [
    # spatial_stats
    "morans_i", "local_morans_i",
    "gearys_c", "local_gearys_c",
    "lees_l", "local_lees_l"
]



"""Functions for interfacing with common tools."""

from .cifti import (
    describe_cifti, extract_cifti_volume, extract_cifti_surface,
    extract_cifti_labels, extract_cifti_surface_labels, deconstruct_cifti
)

from .freesurfer import (
    extract_annot_labels
)

from .gifti import (
    extract_gifti_labels
)

from .surf_parc import (
    vertices_to_parcels, parcels_to_vertices
)

from .interface_utils import (
    PARCIGNORE
)

__all__ = [
    # cifti
    "describe_cifti", "extract_cifti_volume", "extract_cifti_surface",
    "extract_cifti_labels", "extract_cifti_surface_labels", "deconstruct_cifti",
    # freesurfer
    "extract_annot_labels",
    # gifti
    "extract_gifti_labels",
    # surf_parc
    "vertices_to_parcels", "parcels_to_vertices",
    # interface_utils
    "PARCIGNORE"
]



"""Functions for working with FreeSurfer data and parcellations."""

import numpy as np
import nibabel as nib
from netneurotools.interface.interface_utils import PARCIGNORE


def extract_annot_labels(annot_file, parc_ignore=PARCIGNORE):
    """
    Extract vertices and labels from FreeSurfer annotation file.

    Parameters
    ----------
    annot_file : str or Path
        Path to FreeSurfer annotation file.
    parc_ignore : list, optional
        List of labels to ignore.

    Returns
    -------
    surf_data : np.ndarray
        Surface data.
    keys : tuple
        Keys.
    labels : tuple
        Labels.
    """
    surf_data, _, labels = nib.freesurfer.read_annot(annot_file)
    labels = [label.decode() for label in labels]
    keys = np.sort(np.unique(surf_data))
    keys, labels = zip(
        *[
            (key, label)
            for key, label in zip(keys, labels)
            if label not in parc_ignore
        ]
    )
    keys = tuple(map(int, keys))
    return surf_data, keys, labels



"""For testing netneurotools.stats.correlation functionality."""

import pytest
import numpy as np
from netneurotools import stats


@pytest.mark.parametrize('x, y, expected', [
    # basic one-dimensional input
    (range(5), range(5), (1.0, 0.0)),
    # broadcasting occurs regardless of input order
    (np.stack([range(5), range(5, 0, -1)], 1), range(5),
     ([1.0, -1.0], [0.0, 0.0])),
    (range(5), np.stack([range(5), range(5, 0, -1)], 1),
     ([1.0, -1.0], [0.0, 0.0])),
    # correlation between matching columns
    (np.stack([range(5), range(5, 0, -1)], 1),
     np.stack([range(5), range(5, 0, -1)], 1),
     ([1.0, 1.0], [0.0, 0.0]))
])
def test_efficient_pearsonr(x, y, expected):
    """Test efficient_pearsonr function."""
    assert np.allclose(stats.efficient_pearsonr(x, y), expected)


def test_efficient_pearsonr_errors():
    """Test efficient_pearsonr function errors."""
    with pytest.raises(ValueError):
        stats.efficient_pearsonr(range(4), range(5))

    assert all(np.isnan(a) for a in stats.efficient_pearsonr([], []))


@pytest.mark.parametrize(
    "x, y, w, expected",
    [
        (
            np.array([3, 5, 6, 8, 3, 2, 6]),
            np.array([3, 5, 2, 8, 3, 3, 6]),
            np.array([7, 3, 3, 2, 4, 5, 7]),
            0.7356763090950997,
        )
    ],
)
def test_weighted_pearsonr(x, y, w, expected):
    """Test weighted_pearsonr function."""
    assert np.allclose(
        stats.weighted_pearsonr(x, y, w, use_numba=True),
        expected,
    )
    assert np.allclose(
        stats.weighted_pearsonr(x, y, w, use_numba=False),
        expected,
    )


@pytest.mark.parametrize('corr, size, tol, seed', [
    (0.85, (1000,), 0.05, 1234),
    (0.85, (1000, 1000), 0.05, 1234),
    ([[1, 0.5, 0.3], [0.5, 1, 0], [0.3, 0, 1]], (1000,), 0.05, 1234)
])
def test_make_correlated_xy(corr, size, tol, seed):
    """Test make_correlated_xy function."""
    out = stats.make_correlated_xy(corr=corr, size=size,
                                      tol=tol, seed=seed)
    # ensure output is expected shape
    assert out.shape[1:] == size
    assert len(out) == len(corr) if hasattr(corr, '__len__') else 2

    # check outputs are correlated within specified tolerance
    realcorr = np.corrcoef(out.reshape(len(out), -1))
    if len(realcorr) == 2 and not hasattr(corr, '__len__'):
        realcorr = realcorr[0, 1]
    assert np.all(np.abs(realcorr - corr) < tol)

    # check that seed generates reproducible values
    duplicate = stats.make_correlated_xy(corr=corr, size=size,
                                            tol=tol, seed=seed)
    assert np.allclose(out, duplicate)


@pytest.mark.parametrize('corr', [
    (1.5), (-1.5),                                   # outside range of [-1, 1]
    ([0.85]), ([[0.5, 0.5, 0.5], [0.5, 0.5, 0.5]]),  # not 2D / square array
    ([[0.85]]), ([[1, 0.5], [0.5, 0.5]])             # diagonal not equal to 1
])
def test_make_correlated_xy_errors(corr):
    """Test make_correlated_xy function errors."""
    with pytest.raises(ValueError):
        stats.make_correlated_xy(corr)



"""For testing netneurotools.stats.regression functionality."""

import numpy as np
from netneurotools import stats


def test_add_constant():
    """Test adding a constant to a 1D or 2D array."""
    # if provided a vector it will return a 2D array
    assert stats._add_constant(np.random.rand(100)).shape == (100, 2)

    # if provided a 2D array it will return the same, extended by 1 column
    out = stats._add_constant(np.random.rand(100, 100))
    assert out.shape == (100, 101) and np.all(out[:, -1] == 1)
